# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base DatasetBuilderTestCase to test a DatasetBuilder."""

import collections
import contextlib
import difflib
import hashlib
import itertools
import numbers
import os
import textwrap
from typing import Iterator, List, Mapping, Optional, Union
from unittest import mock

from absl import logging
from absl.testing import parameterized
from etils import epath
import numpy as np
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import dataset_metadata
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import download
from tensorflow_datasets.core import load
from tensorflow_datasets.core import utils
from tensorflow_datasets.core import visibility
from tensorflow_datasets.core.download import checksums
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.testing import feature_test_case
from tensorflow_datasets.testing import test_utils

# `os` module Functions for which tf.io.gfile equivalent should be preferred.
FORBIDDEN_OS_FUNCTIONS = (
    "chmod",
    "chown",
    "link",
    "listdir",
    "lstat",
    "makedirs",
    "mkdir",
    "mknod",
    "open",
    "pathconf",
    "readlink",
    "remove",
    "removedirs",
    "rename",
    "renames",
    "rmdir",
    "stat",
    "statvfs",
    "symlink",
    "unlink",
    "walk",
)
FORBIDDEN_OS_PATH_FUNCTIONS = (
    "exists",
    "isdir",
    "isfile",
)

_ORGINAL_NP_LOAD = np.load


def _np_load(file_, mmap_mode=None, allow_pickle=False, **kwargs):
  if not hasattr(file_, "read"):
    raise AssertionError(
        "You MUST pass a `tf.io.gfile.GFile` or file-like object to `np.load`."
    )
  if allow_pickle:
    raise AssertionError("Unpicling files is forbidden for security reasons.")
  return _ORGINAL_NP_LOAD(file_, mmap_mode, allow_pickle, **kwargs)


class DatasetBuilderTestCase(
    parameterized.TestCase, feature_test_case.SubTestCase
):
  """Inherit this class to test your DatasetBuilder class.

  You must set the following class attributes:

    * DATASET_CLASS: class object of DatasetBuilder you want to test.
    * SPLITS: dict mapping split name to expected number of examples.

  You may set the following class attributes:

    * VERSION: `str`. The version used to run the test. eg: '1.2.*'.
      Defaults to None (canonical version).
    * BUILDER_CONFIG_NAMES_TO_TEST: `list[str | tfds.core.BuilderConfig]`,
      the list of builder configs that should be tested. If None, all the
      BUILDER_CONFIGS from the class will be tested.
    * DL_EXTRACT_RESULT: `dict[str, str]`, the returned result of mocked
      `download_and_extract` method. The values should be the path of files
      present in the `fake_examples` directory, relative to that directory.
      If not specified, path to `fake_examples` will always be returned.
    * DL_EXTRACT_ONLY_RESULT: `dict[str, str]`, the returned result of mocked
      `extract` method. The values should be the path of files present in the
      `fake_examples` directory, relative to that directory. If not specified:
      will call DownloadManager `extract` method.
    * DL_DOWNLOAD_RESULT: `dict[str, str]`, the returned result of mocked
      `download_and_extract` method. The values should be the path of files
      present in the `fake_examples` directory, relative to that directory.
      If not specified: will use DL_EXTRACT_RESULT (this is due to backwards
      compatibility and will be removed in the future).
    * EXAMPLE_DIR: `str`, the base directory in in which fake examples are
      contained. Optional; defaults to `<dataset dir>/dummy_data/`.
    * OVERLAPPING_SPLITS: `list[str]`, splits containing examples from other
      splits (e.g. a "example" split containing pictures from other splits).
    * MOCK_OUT_FORBIDDEN_OS_FUNCTIONS: `bool`, defaults to True. Set to False to
      disable checks preventing usage of `os` or builtin functions instead of
      recommended `tf.io.gfile` API.
    * SKIP_CHECKSUMS: Checks that the urls called by `dl_manager.download`
      are registered.
    * SKIP_TF1_GRAPH_MODE: Runs in eager mode only.

  This test case will check for the following:

   - the dataset builder is correctly registered, i.e. `tfds.load(name)` works;
   - the dataset builder can read the fake examples stored in
       testing/test_data/fake_examples/{dataset_name};
   - the dataset builder can produce serialized data;
   - the dataset builder produces a valid Dataset object from serialized data
     - in eager mode;
     - in graph mode.
   - the produced Dataset examples have the expected dimensions and types;
   - the produced Dataset has and the expected number of examples;
   - a example is not part of two splits, or one of these splits is whitelisted
       in OVERLAPPING_SPLITS.
  """

  DATASET_CLASS = None
  SPLITS = {}
  VERSION = None
  BUILDER_CONFIG_NAMES_TO_TEST: Optional[
      List[Union[str, dataset_builder.BuilderConfig]]
  ] = None
  DL_EXTRACT_RESULT: Optional[str] = None
  DL_EXTRACT_ONLY_RESULT: Optional[str] = None
  DL_DOWNLOAD_RESULT: Optional[str] = None
  EXAMPLE_DIR = None
  OVERLAPPING_SPLITS = []
  MOCK_OUT_FORBIDDEN_OS_FUNCTIONS = True
  SKIP_CHECKSUMS = False
  SKIP_TF1_GRAPH_MODE = True

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    name = cls.__name__
    # Check class has the right attributes
    if cls.DATASET_CLASS is None or not callable(cls.DATASET_CLASS):
      raise AssertionError(
          "Assign your DatasetBuilder class to %s.DATASET_CLASS." % name
      )

    cls._available_cm = visibility.set_availables_tmp(
        [
            visibility.DatasetType.TFDS_PUBLIC,
        ]
    )
    cls._available_cm.__enter__()  # pylint: disable=protected-access

  @classmethod
  def tearDownClass(cls):
    super().tearDownClass()
    cls._available_cm.__exit__(None, None, None)  # pylint: disable=protected-access

  def setUp(self):
    super(DatasetBuilderTestCase, self).setUp()
    self.patchers = []
    self.builder = self._make_builder()

    if self.MOCK_OUT_FORBIDDEN_OS_FUNCTIONS:
      self._mock_out_forbidden_os_functions()

    # Track the urls which are downloaded to validate the checksums
    # The `dl_manager.download` and `dl_manager.download_and_extract` are
    # patched to record the urls in `_download_urls`.
    # Calling `dl_manager.download_checksums` stop the url
    # registration (as checksums are stored remotelly)
    # `_test_checksums` validates the recorded urls.
    self._download_urls = set()
    self._stop_record_download = False

  def tearDown(self):
    super(DatasetBuilderTestCase, self).tearDown()
    for patcher in self.patchers:
      patcher.stop()

  @utils.classproperty
  @classmethod
  @utils.memoize()
  def dummy_data(cls) -> epath.Path:  # pylint: disable=no-self-argument
    """Path to the `dummy_data/` directory."""
    if cls is DatasetBuilderTestCase:  # Required for build_api_docs
      return None  # pytype: disable=bad-return-type

    dummy_data_expected = cls.DATASET_CLASS.code_path.parent / "dummy_data"
    fake_example_dir = epath.Path(test_utils.fake_examples_dir())
    if cls.EXAMPLE_DIR is not None:
      dummy_data_found = epath.Path(cls.EXAMPLE_DIR)
      dummy_data_expected = dummy_data_found  # Dir to display in the error
    elif dummy_data_expected.exists():
      dummy_data_found = dummy_data_expected
    else:
      dummy_data_found = fake_example_dir / cls.DATASET_CLASS.name

    if not dummy_data_found.exists():
      err_msg = f"Dummy data not found in: {dummy_data_expected}"
      raise ValueError(err_msg)
    return dummy_data_found

  def _mock_out_forbidden_os_functions(self):
    """Raises error if forbidden os functions are called instead of gfile."""
    err = AssertionError(
        "Do not use `os`, but `tf.io.gfile` module instead. "
        "This makes code compatible with more filesystems."
    )
    sep = os.path.sep
    mock_os_path = mock.Mock(os.path, wraps=os.path)
    mock_os_path.sep = sep
    for fop in FORBIDDEN_OS_PATH_FUNCTIONS:
      getattr(mock_os_path, fop).side_effect = err
    mock_os = mock.Mock(os, path=mock_os_path, fspath=os.fspath)
    for fop in FORBIDDEN_OS_FUNCTIONS:
      if os.name == "nt" and not hasattr(os, fop):
        continue  # Not all `os` functions are available on Windows (ex: chmod).
      getattr(mock_os, fop).side_effect = err
    os_patcher = mock.patch(
        self.DATASET_CLASS.__module__ + ".os", mock_os, create=True
    )
    os_patcher.start()
    self.patchers.append(os_patcher)

    mock_builtins = __builtins__.copy()  # pytype: disable=module-attr
    mock_builtins["open"] = mock.Mock(side_effect=err)
    open_patcher = mock.patch(
        self.DATASET_CLASS.__module__ + ".__builtins__", mock_builtins
    )
    open_patcher.start()
    self.patchers.append(open_patcher)

    # It's hard to mock open within numpy, so mock np.load.
    np_load_patcher = mock.patch("numpy.load", _np_load)
    np_load_patcher.start()
    self.patchers.append(np_load_patcher)

  def test_baseclass(self):
    self.assertIsInstance(
        self.builder,
        dataset_builder.DatasetBuilder,
        "Dataset class must inherit from `dataset_builder.DatasetBuilder`.",
    )
    # Since class was instantiated and base class is ABCMeta, then we know
    # all needed methods were implemented.

  def test_registered(self):
    self.assertIn(
        self.builder.name,
        load.list_builders(
            with_community_datasets=False,
        ),
    )

  def test_info(self):
    info = self.builder.info
    self.assertIsInstance(info, dataset_info.DatasetInfo)
    self.assertEqual(self.builder.name, info.name)

  def test_tags_are_valid(self):
    metadata = self.builder.get_metadata()
    tags = set(metadata.tags)
    valid_tags = set(dataset_metadata.valid_tags())
    invalid_tags = tags - valid_tags
    self.assertEmpty(invalid_tags, "There are invalid tags.")

  def _add_url(self, url_or_urls):
    if self._stop_record_download:
      # Stop record the checksums if dl_manager.download_checksums has been
      # called (as checksums may be stored remotelly)
      return
    if isinstance(url_or_urls, download.resource.Resource):
      self._download_urls.add(url_or_urls.url)
    else:
      self._download_urls.add(url_or_urls)

  def _prepare_download_results(
      self,
      downloaded_files: Union[Mapping[str, str], str],
  ) -> Mapping[str, epath.PathLike]:
    return tf.nest.map_structure(
        lambda fname: self.dummy_data / fname,
        downloaded_files,
    )

  def _get_dl_extract_result(self, url):
    tf.nest.map_structure(self._add_url, url)
    del url
    if self.DL_EXTRACT_RESULT is None:
      return self.dummy_data
    return self._prepare_download_results(self.DL_EXTRACT_RESULT)

  def _get_dl_extract_only_result(self, url):
    if self.DL_EXTRACT_ONLY_RESULT:
      tf.nest.map_structure(self._add_url, url)
      return self._prepare_download_results(self.DL_EXTRACT_ONLY_RESULT)

  def _get_dl_download_result(self, url):
    tf.nest.map_structure(self._add_url, url)
    if self.DL_DOWNLOAD_RESULT is None:
      # This is only to be backwards compatible with old approach.
      # In the future it will be replaced with using self.dummy_data.
      return self._get_dl_extract_result(url)
    return self._prepare_download_results(self.DL_DOWNLOAD_RESULT)

  def _get_dl_download_kaggle_result(self, competition_or_dataset):
    del competition_or_dataset  # Unused
    return self._prepare_download_results(self.DL_DOWNLOAD_RESULT)

  def _download_checksums(self, url):
    del url  # Unused
    self._stop_record_download = True

  @property
  def dataset_class(self):
    if self.DATASET_CLASS is None or not issubclass(
        self.DATASET_CLASS, dataset_builder.DatasetBuilder
    ):
      raise ValueError(
          "DATASET_CLASS is not a subclass of DatasetBuilder. Call setUpClass()"
          " properly."
      )
    return self.DATASET_CLASS

  def _make_builder(self, config=None):
    return self.dataset_class(  # pylint: disable=not-callable
        data_dir=self.tmp_dir, config=config, version=self.VERSION
    )

  @test_utils.run_in_graph_and_eager_modes()
  def test_download_and_prepare_as_dataset(self):
    if not tf.executing_eagerly() and self.SKIP_TF1_GRAPH_MODE:
      logging.warning("Skipping tests in non-eager mode")
      return

    # Extract configs to test
    configs_to_test: List[Union[str, dataset_builder.BuilderConfig]] = []
    if self.BUILDER_CONFIG_NAMES_TO_TEST:
      for config in self.BUILDER_CONFIG_NAMES_TO_TEST:  # pylint: disable=not-an-iterable
        if isinstance(config, dataset_builder.BuilderConfig):
          configs_to_test.append(config)
        elif config in self.builder.builder_configs:
          # Append the `name` rather than the config due to
          # https://github.com/tensorflow/datasets/issues/2348
          configs_to_test.append(config)
        else:
          raise ValueError(
              f"Invalid config {config} specified in test."
              f"Available: {list(self.builder.builder_configs)}"
          )
    else:
      configs_to_test.extend(cfg.name for cfg in self.builder.BUILDER_CONFIGS)

    print(f"Total configs: {len(configs_to_test)}")
    if configs_to_test:
      for config in configs_to_test:
        config_name = config if isinstance(config, str) else config.name
        with self._subTest(config_name):
          print(f"Testing config {config_name}")
          builder = self._make_builder(config=config)
          self._download_and_prepare_as_dataset(builder)
    else:
      self._download_and_prepare_as_dataset(self.builder)

    if not self.SKIP_CHECKSUMS:
      with self._subTest("url_checksums"):
        self._test_checksums()

  def _test_checksums(self):
    # If no call to `dl_manager.download`, then no need to check url presence.
    if not self._download_urls:
      return

    err_msg = (
        "Did you forget to record checksums with `--register_checksums` ? See"
        " instructions at:"
        " https://www.tensorflow.org/datasets/add_dataset#run_the_generation_codeIf"
        " want to opt-out of checksums validation, please add `SKIP_CHECKSUMS ="
        " True` to the `DatasetBuilderTestCase`.\n"
    )
    url_infos = self.dataset_class.url_infos
    filepath = self.dataset_class._checksums_path  # pylint: disable=protected-access
    # Legacy checksums: Search in `checksums/` dir
    if url_infos is None:
      legacy_filepath = checksums._checksum_paths().get(self.builder.name)  # pylint: disable=protected-access
      if legacy_filepath and legacy_filepath.exists():
        filepath = legacy_filepath
        url_infos = checksums.load_url_infos(filepath)
    # Checksums not present neither in legacy nor package
    if url_infos is None:
      raise FileNotFoundError(
          f"Checksums file not found at: {filepath}\n"
          f"{err_msg}\n"
      )

    missing_urls = self._download_urls - set(url_infos.keys())
    self.assertEmpty(
        missing_urls,
        f"Some urls checksums are missing at: {filepath}\n{err_msg}",
    )

  def _download_and_prepare_as_dataset(self, builder):
    # Provide the manual dir only if builder has MANUAL_DOWNLOAD_INSTRUCTIONS
    # set.

    missing_dir_mock = mock.PropertyMock(
        side_effect=Exception("Missing MANUAL_DOWNLOAD_INSTRUCTIONS")
    )

    manual_dir = (
        self.dummy_data
        if builder.MANUAL_DOWNLOAD_INSTRUCTIONS
        else missing_dir_mock
    )

    patches = {
        "download_and_extract": self._get_dl_extract_result,
        "download": self._get_dl_download_result,
        "download_kaggle_data": self._get_dl_download_kaggle_result,
        "download_checksums": self._download_checksums,
        "manual_dir": manual_dir,
        "download_dir": self.dummy_data,
    }
    if self.DL_EXTRACT_ONLY_RESULT:
      patches["extract"] = self._get_dl_extract_only_result

    with mock.patch.multiple(
        "tensorflow_datasets.core.download.DownloadManager", **patches
    ):
      # For Beam datasets, set-up the runner config
      beam_runner = None

      download_config = download.DownloadConfig(
          compute_stats=download.ComputeStatsMode.SKIP,
          beam_runner=beam_runner,
      )
      with self._test_key_not_local_path(builder):
        builder.download_and_prepare(download_config=download_config)

    with self._subTest("as_dataset"):
      self._assertAsDataset(builder)

    with self._subTest("num_examples"):
      self._assertNumSamples(builder)

    with self._subTest("reload"):
      # When reloading the dataset, metadata should been reloaded too.

      builder_reloaded = self._make_builder(config=builder.builder_config)
      self._assertNumSamples(builder_reloaded)

      # After reloading, as_dataset should still be working
      with self._subTest("as_dataset"):
        self._assertAsDataset(builder_reloaded)

    with self._subTest("config_description"):
      self._test_description_builder_config(builder)

    with self._subTest("default_config"):
      self._test_default_builder_config(builder)

  @contextlib.contextmanager
  def _test_key_not_local_path(self, builder) -> Iterator[None]:
    if not isinstance(builder, dataset_builder.GeneratorBasedBuilder):
      yield  # Do not test non-generator builder
      return

    original_generate_examples = builder._generate_examples  # pylint: disable=protected-access

    def _iter_examples(generator):
      for key, ex in generator:
        self._assert_key_valid(key)
        yield key, ex

    def new_generate_examples(*args, **kwargs):
      examples = original_generate_examples(*args, **kwargs)
      try:
        import apache_beam as beam  # pylint: disable=g-import-not-at-top
      except ImportError:
        beam = None
      if beam and isinstance(examples, (beam.PCollection, beam.PTransform)):
        return examples  # Beam keys not supported for now
      elif isinstance(examples, collections.abc.Iterable):
        return _iter_examples(examples)
      else:  # Unexpected
        return examples

    with mock.patch.object(
        builder, "_generate_examples", new_generate_examples
    ):
      yield

  def _assert_key_valid(self, key):
    if isinstance(key, str) and os.fspath(self.dummy_data) in key:
      err_msg = (
          "Key yield in '_generate_examples' method "
          f"contain user directory path: {key}.\n"
          "This makes the dataset example order non-deterministic. "
          "Please use `filepath.name`, or `os.path.basename` instead."
      )
      raise ValueError(err_msg)

  def _assertAsDataset(self, builder):
    split_to_checksums = {}  # {"split": set(examples_checksums)}
    for split_name, expected_examples_number in self.SPLITS.items():
      ds = builder.as_dataset(split=split_name)
      spec = tf.data.DatasetSpec.from_value(ds)
      compare_shapes_and_types(
          builder.info.features.get_tensor_info(),
          # We use _element_spec because element_spec was added in TF2.5+.
          element_spec=spec._element_spec,  # pylint: disable=protected-access
      )
      examples = list(
          dataset_utils.as_numpy(builder.as_dataset(split=split_name))
      )
      split_to_checksums[split_name] = set(checksum(rec) for rec in examples)
      self.assertLen(examples, expected_examples_number)
    for (split1, hashes1), (split2, hashes2) in itertools.combinations(
        split_to_checksums.items(), 2
    ):
      if split1 in self.OVERLAPPING_SPLITS or split2 in self.OVERLAPPING_SPLITS:
        continue
      self.assertFalse(
          hashes1.intersection(hashes2),
          "Splits '%s' and '%s' are overlapping. Are you sure you want to "
          "have the same objects in those splits? If yes, add one one of "
          "them to OVERLAPPING_SPLITS class attribute." % (split1, split2),
      )

  def _assertNumSamples(self, builder):
    for split_name, expected_num_examples in self.SPLITS.items():
      self.assertEqual(
          builder.info.splits[split_name].num_examples,
          expected_num_examples,
          (
              f"Number of examples in split '{split_name}' "
              "do not match what is expected."
          ),
      )
    self.assertEqual(
        builder.info.splits.total_num_examples,
        sum(self.SPLITS.values()),
    )

  def _test_description_builder_config(self, builder):
    # Do not test on external datasets for backward compatibility
    # TODO(tfds): Update once datasets are migrated
    if "tensorflow_datasets" not in self.DATASET_CLASS.__module__:
      return

    # If configs specified, ensure they are all valid
    if builder.builder_config and builder.builder_config.description:
      err_msg = textwrap.dedent(
          """\
          The BuilderConfig description should be a one-line description of
          the config.
          It shouldn't be the same as `builder.info.description` to avoid
          redundancy. Both `config.description` and `builder.info.description`
          will be displayed in the catalog.
          """
      )
      ratio = difflib.SequenceMatcher(
          None,
          builder.builder_config.description,
          builder.info.description,
      ).ratio()
      if ratio > 0.9:
        raise AssertionError(err_msg)

  def _test_default_builder_config(self, builder):
    if not builder.DEFAULT_BUILDER_CONFIG_NAME:
      return
    builder_config_names = {config.name for config in builder.BUILDER_CONFIGS}
    self.assertIn(builder.DEFAULT_BUILDER_CONFIG_NAME, builder_config_names)


def checksum(example):
  """Computes the md5 for a given example."""

  def _bytes_flatten(flat_str, element):
    """Recursively flatten an element to its byte representation."""
    if isinstance(element, numbers.Number):
      # In python3, bytes(-3) is not allowed (or large numbers),
      # so convert to str to avoid problems.
      element = str(element)
    if isinstance(element, dict):
      for k, v in sorted(element.items()):
        flat_str.append(k)
        _bytes_flatten(flat_str, v)
    elif isinstance(element, str):
      if hasattr(element, "decode"):
        # Python2 considers bytes to be str, but are almost always latin-1
        # encoded bytes here. Extra step needed to avoid DecodeError.
        element = element.decode("latin-1")
      flat_str.append(element)
    elif isinstance(
        element, (tf.RaggedTensor, tf.compat.v1.ragged.RaggedTensorValue)
    ):
      flat_str.append(str(element.to_list()))
    elif isinstance(element, (np.ndarray, np.generic)):
      # tf.Tensor() returns np.array of dtype object, which don't work
      # with x.to_bytes(). So instead convert numpy into list.
      if element.dtype.type is np.object_:
        flat_str.append(str(tuple(element.shape)))
        flat_str.append(str(list(element.ravel())))
      else:
        flat_str.append(element.tobytes())
    elif isinstance(element, dataset_utils._IterableDataset):  # pylint: disable=protected-access
      for nested_e in element:
        _bytes_flatten(flat_str, nested_e)
    else:
      flat_str.append(bytes(element))
    return flat_str

  flat_str = _bytes_flatten([], example)
  flat_bytes = [
      s.encode("utf-8") if not isinstance(s, bytes) else s for s in flat_str
  ]
  flat_bytes = b"".join(flat_bytes)

  hash_ = hashlib.md5()
  hash_.update(flat_bytes)
  return hash_.hexdigest()


def compare_shapes_and_types(tensor_info, element_spec):
  """Compare shapes and types between TensorInfo and Dataset types/shapes."""
  for feature_name, (feature_info, spec) in utils.zip_dict(
      tensor_info, element_spec
  ):
    if isinstance(spec, tf.data.DatasetSpec):
      # We use _element_spec because element_spec was added in TF2.5+.
      compare_shapes_and_types(feature_info, spec._element_spec)  # pylint: disable=protected-access
    elif isinstance(feature_info, dict):
      compare_shapes_and_types(feature_info, spec)
    else:
      # Some earlier versions of TF don't expose dtype and shape for the
      # RaggedTensorSpec, so we use the protected versions.
      if feature_info.dtype != spec._dtype:  # pylint: disable=protected-access
        raise TypeError(
            f"Feature {feature_name} has type {feature_info} but expected"
            f" {spec}"
        )
      utils.assert_tf_shape_match(
          tf.TensorShape(feature_info.shape), spec._shape
      )  # pylint: disable=protected-access



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""scrolls dataset."""
import dataclasses
import json

from etils import epath
import tensorflow_datasets.public_api as tfds

# Corresponding HF dataset:
# https://huggingface.co/datasets/tau/scrolls/blob/main/scrolls.py

_SCROLLS_CITATION = """
@misc{shaham2022scrolls,
      title={SCROLLS: Standardized CompaRison Over Long Language Sequences},
      author={Uri Shaham and Elad Segal and Maor Ivgi and Avia Efrat and Ori Yoran and Adi Haviv and Ankit Gupta and Wenhan Xiong and Mor Geva and Jonathan Berant and Omer Levy},
      year={2022},
      eprint={2201.03533},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
Note that each SCROLLS dataset has its own citation. Please see the source to
get the correct citation for each contained dataset.
"""

_SCROLLS_DESCRIPTION = """
SCROLLS: Standardized CompaRison Over Long Language Sequences.
A suite of natural language tfds.core.that require reasoning over long texts.
https://scrolls-benchmark.com/
"""

_SUMM_SCREEN_CITATION = r"""
@misc{chen2021summscreen,
      title={SummScreen: A Dataset for Abstractive Screenplay Summarization},
      author={Mingda Chen and Zewei Chu and Sam Wiseman and Kevin Gimpel},
      year={2021},
      eprint={2104.07091},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}"""

_QASPER_CITATION = r"""
@inproceedings{dasigi-etal-2021-dataset,
    title = "A Dataset of Information-Seeking Questions and Answers Anchored in Research Papers",
    author = "Dasigi, Pradeep  and
      Lo, Kyle  and
      Beltagy, Iz  and
      Cohan, Arman  and
      Smith, Noah A.  and
      Gardner, Matt",
    booktitle = "Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies",
    month = jun,
    year = "2021",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2021.naacl-main.365",
    doi = "10.18653/v1/2021.naacl-main.365",
    pages = "4599--4610",
    abstract = "Readers of academic research papers often read with the goal of answering specific questions. Question Answering systems that can answer those questions can make consumption of the content much more efficient. However, building such tools requires data that reflect the difficulty of the task arising from complex reasoning about claims made in multiple parts of a paper. In contrast, existing information-seeking question answering tfds.core.usually contain questions about generic factoid-type information. We therefore present Qasper, a dataset of 5049 questions over 1585 Natural Language Processing papers. Each question is written by an NLP practitioner who read only the title and abstract of the corresponding paper, and the question seeks information present in the full text. The questions are then answered by a separate set of NLP practitioners who also provide supporting evidence to answers. We find that existing models that do well on other QA tasks do not perform well on answering these questions, underperforming humans by at least 27 F1 points when answering them from entire papers, motivating further research in document-grounded, information-seeking QA, which our dataset is designed to facilitate.",
}"""

_QMSUM_CITATION = r"""@inproceedings{zhong-etal-2021-qmsum,
    title = "{QMS}um: A New Benchmark for Query-based Multi-domain Meeting Summarization",
    author = "Zhong, Ming  and
      Yin, Da  and
      Yu, Tao  and
      Zaidi, Ahmad  and
      Mutuma, Mutethia  and
      Jha, Rahul  and
      Awadallah, Ahmed Hassan  and
      Celikyilmaz, Asli  and
      Liu, Yang  and
      Qiu, Xipeng  and
      Radev, Dragomir",
    booktitle = "Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies",
    month = jun,
    year = "2021",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2021.naacl-main.472",
    doi = "10.18653/v1/2021.naacl-main.472",
    pages = "5905--5921",
    abstract = "Meetings are a key component of human collaboration. As increasing numbers of meetings are recorded and transcribed, meeting summaries have become essential to remind those who may or may not have attended the meetings about the key decisions made and the tasks to be completed. However, it is hard to create a single short summary that covers all the content of a long meeting involving multiple people and topics. In order to satisfy the needs of different types of users, we define a new query-based multi-domain meeting summarization task, where models have to select and summarize relevant spans of meetings in response to a query, and we introduce QMSum, a new benchmark for this task. QMSum consists of 1,808 query-summary pairs over 232 meetings in multiple domains. Besides, we investigate a locate-then-summarize method and evaluate a set of strong summarization baselines on the task. Experimental results and manual analysis reveal that QMSum presents significant challenges in long meeting summarization for future research. Dataset is available at \url{https://github.com/Yale-LILY/QMSum}.",
}"""

_NARRATIVE_QA_CITATION = r"""
@article{kovcisky2018narrativeqa,
  title={The narrativeqa reading comprehension challenge},
  author={Ko{\v{c}}isk{\'y}, Tom{\'a}{\v{s}} and Schwarz, Jonathan and Blunsom, Phil and Dyer, Chris and Hermann, Karl Moritz and Melis, G{\'a}bor and Grefenstette, Edward},
  journal={Transactions of the Association for Computational Linguistics},
  volume={6},
  pages={317--328},
  year={2018},
  publisher={MIT Press}
}"""

_GOV_REPORT_CITATION = r"""
@inproceedings{huang-etal-2021-efficient,
    title = "Efficient Attentions for Long Document Summarization",
    author = "Huang, Luyang  and
      Cao, Shuyang  and
      Parulian, Nikolaus  and
      Ji, Heng  and
      Wang, Lu",
    booktitle = "Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies",
    month = jun,
    year = "2021",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2021.naacl-main.112",
    doi = "10.18653/v1/2021.naacl-main.112",
    pages = "1419--1436",
    abstract = "The quadratic computational and memory complexities of large Transformers have limited their scalability for long document summarization. In this paper, we propose Hepos, a novel efficient encoder-decoder attention with head-wise positional strides to effectively pinpoint salient information from the source. We further conduct a systematic study of existing efficient self-attentions. Combined with Hepos, we are able to process ten times more tokens than existing models that use full attentions. For evaluation, we present a new dataset, GovReport, with significantly longer documents and summaries. Results show that our models produce significantly higher ROUGE scores than competitive comparisons, including new state-of-the-art results on PubMed. Human evaluation also shows that our models generate more informative summaries with fewer unfaithful errors.",
}"""

_CONTRACT_NLI_CITATION = """\
@inproceedings{koreeda-manning-2021-contractnli,
    title = "ContractNLI: A Dataset for Document-level Natural Language Inference for Contracts",
    author = "Koreeda, Yuta  and
      Manning, Christopher D.",
    booktitle = "Findings of the Association for Computational Linguistics: EMNLP 2021",
    year = "2021",
    publisher = "Association for Computational Linguistics"
}
"""
_QUALITY_CITATION = """\
@article{pang2021quality,
  title={{QuALITY}: Question Answering with Long Input Texts, Yes!},
  author={Pang, Richard Yuanzhe and Parrish, Alicia and Joshi, Nitish and Nangia, Nikita and Phang, Jason and Chen, Angelica and Padmakumar, Vishakh and Ma, Johnny and Thompson, Jana and He, He and Bowman, Samuel R.},
  journal={arXiv preprint arXiv:2112.08608},
  year={2021}
}
"""


@dataclasses.dataclass
class ScrollsConfig(tfds.core.BuilderConfig):
  """BuilderConfig for SCROLLS."""

  data_url: str = ""
  citation: str = ""
  url: str = ""


_INPUT_KEY = "input"
_OUTPUT_KEY = "output"
_FEATURES = ("id", "pid", _INPUT_KEY, _OUTPUT_KEY)


class Scrolls(tfds.core.GeneratorBasedBuilder):
  """The SCROLLS benchmark."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  BUILDER_CONFIGS = [
      ScrollsConfig(
          name="summ_screen_fd",
          description="summ_screen_fd subset",
          data_url="https://scrolls-tau.s3.us-east-2.amazonaws.com/summ_screen_fd.zip",
          citation=_SUMM_SCREEN_CITATION,
          url="https://github.com/mingdachen/SummScreen",
      ),
      ScrollsConfig(
          name="qasper",
          description="qasper subset",
          data_url="https://scrolls-tau.s3.us-east-2.amazonaws.com/qasper.zip",
          citation=_QASPER_CITATION,
          url="https://allenai.org/project/qasper",
      ),
      ScrollsConfig(
          name="qmsum",
          description="qmsum subset",
          data_url="https://scrolls-tau.s3.us-east-2.amazonaws.com/qmsum.zip",
          citation=_QMSUM_CITATION,
          url="https://github.com/Yale-LILY/QMSum",
      ),
      ScrollsConfig(
          name="narrative_qa",
          description="narrative_qa subset",
          data_url=(
              "https://scrolls-tau.s3.us-east-2.amazonaws.com/narrative_qa.zip"
          ),
          citation=_NARRATIVE_QA_CITATION,
          url="https://deepmind.com/research/publications/narrativeqa-reading-comprehension-challenge",
      ),
      ScrollsConfig(
          name="gov_report",
          description="gov_report subset",
          data_url=(
              "https://scrolls-tau.s3.us-east-2.amazonaws.com/gov_report.zip"
          ),
          citation=_GOV_REPORT_CITATION,
          url="https://gov-report-data.github.io/",
      ),
      ScrollsConfig(
          name="contract_nli",
          description="contract_nli subset",
          data_url=(
              "https://scrolls-tau.s3.us-east-2.amazonaws.com/contract_nli.zip"
          ),
          citation=_CONTRACT_NLI_CITATION,
          url="https://stanfordnlp.github.io/contract-nli/",
      ),
      ScrollsConfig(
          name="quality",
          description="quality subset",
          data_url="https://scrolls-tau.s3.us-east-2.amazonaws.com/quality.zip",
          citation=_QUALITY_CITATION,
          url="https://github.com/nyu-mll/quality",
      ),
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_SCROLLS_DESCRIPTION + self.builder_config.description,
        features=tfds.features.FeaturesDict(
            {feature: tfds.features.Text() for feature in _FEATURES}
        ),
        supervised_keys=(_INPUT_KEY, _OUTPUT_KEY),
        homepage=self.builder_config.url,
        citation=self.builder_config.citation + "\n" + _SCROLLS_CITATION,
    )

  def _split_generators(self, dl_manager):
    dl_dir = dl_manager.download_and_extract(self.builder_config.data_url)
    task_name = task_name = _get_task_name_from_data_url(
        self.builder_config.data_url
    )

    return {
        "train": self._generate_examples(dl_dir / task_name / "train.jsonl"),
        "validation": self._generate_examples(
            dl_dir / task_name / "validation.jsonl"
        ),
        "test": self._generate_examples(dl_dir / task_name / "test.jsonl"),
    }

  def _generate_examples(self, path):
    with epath.Path(path).open() as f:
      for line in f:
        row = json.loads(line)

        # Test set has 'null' for 'output'
        yield row["pid"], {
            k: row[k] if row[k] is not None else "" for k in _FEATURES
        }


def _get_task_name_from_data_url(data_url):
  return data_url.split("/")[-1].split(".")[0]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Corrupted Cifar10 Dataset.

Cifar10Corrupted is a dataset generated by adding 15 common corruptions + 4
extra corruptions to the test images in the Cifar10 dataset. This dataset wraps
the corrupted Cifar10 test images uploaded by the original authors.
"""

import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """\
Cifar10Corrupted is a dataset generated by adding 15 common corruptions + 4
extra corruptions to the test images in the Cifar10 dataset. This dataset wraps
the corrupted Cifar10 test images uploaded by the original authors.
"""

_CITATION = """\
@inproceedings{
  hendrycks2018benchmarking,
  title={Benchmarking Neural Network Robustness to Common Corruptions and Perturbations},
  author={Dan Hendrycks and Thomas Dietterich},
  booktitle={International Conference on Learning Representations},
  year={2019},
  url={https://openreview.net/forum?id=HJz6tiCqYm},
}
"""

_CIFAR_IMAGE_SIZE = (32, 32, 3)
_CIFAR_CLASSES = 10
_DOWNLOAD_URL = 'https://zenodo.org/record/2535967/files/CIFAR-10-C.tar'
_CORRUPTIONS_TO_FILENAMES = {
    'gaussian_noise': 'gaussian_noise.npy',
    'shot_noise': 'shot_noise.npy',
    'impulse_noise': 'impulse_noise.npy',
    'defocus_blur': 'defocus_blur.npy',
    'frosted_glass_blur': 'glass_blur.npy',
    'motion_blur': 'motion_blur.npy',
    'zoom_blur': 'zoom_blur.npy',
    'snow': 'snow.npy',
    'frost': 'frost.npy',
    'fog': 'fog.npy',
    'brightness': 'brightness.npy',
    'contrast': 'contrast.npy',
    'elastic': 'elastic_transform.npy',
    'pixelate': 'pixelate.npy',
    'jpeg_compression': 'jpeg_compression.npy',
    'gaussian_blur': 'gaussian_blur.npy',
    'saturate': 'saturate.npy',
    'spatter': 'spatter.npy',
    'speckle_noise': 'speckle_noise.npy',
}
_CORRUPTIONS, _FILENAMES = zip(*sorted(_CORRUPTIONS_TO_FILENAMES.items()))
_DIRNAME = 'CIFAR-10-C'
_LABELS_FILENAME = 'labels.npy'

BENCHMARK_CORRUPTIONS = [
    'gaussian_noise',
    'shot_noise',
    'impulse_noise',
    'defocus_blur',
    'frosted_glass_blur',
    'motion_blur',
    'zoom_blur',
    'snow',
    'frost',
    'fog',
    'brightness',
    'contrast',
    'elastic',
    'pixelate',
    'jpeg_compression',
]

EXTRA_CORRUPTIONS = [
    'gaussian_blur',
    'saturate',
    'spatter',
    'speckle_noise',
]


class Cifar10CorruptedConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Cifar10Corrupted."""

  def __init__(self, *, corruption_type, severity, **kwargs):
    """Constructor.

    Args:
      corruption_type: string, must be one of the items in _CORRUPTIONS.
      severity: integer, bewteen 1 and 5.
      **kwargs: keyword arguments forwarded to super.
    """
    super(Cifar10CorruptedConfig, self).__init__(**kwargs)
    self.corruption = corruption_type
    self.severity = severity


def _make_builder_configs():
  """Construct a list of BuilderConfigs.

  Construct a list of 95 Cifar10CorruptedConfig objects, corresponding to
  the 15 corruption types + 4 extra corruptions and 5 severities.

  Returns:
    A list of 95 Cifar10CorruptedConfig objects.
  """
  config_list = []
  for corruption in _CORRUPTIONS:
    for severity in range(1, 6):
      config_list.append(
          Cifar10CorruptedConfig(
              name=corruption + '_' + str(severity),
              description='Corruption method: '
              + corruption
              + ', severity level: '
              + str(severity),
              corruption_type=corruption,
              severity=severity,
          )
      )
  return config_list


class Cifar10Corrupted(tfds.core.GeneratorBasedBuilder):
  """Corrupted Cifar10 dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'New split API (https://tensorflow.org/datasets/splits)',
  }
  BUILDER_CONFIGS = _make_builder_configs()

  def _info(self):
    """Returns basic information of dataset.

    Returns:
      tfds.core.DatasetInfo.
    """
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(shape=_CIFAR_IMAGE_SIZE),
            'label': tfds.features.ClassLabel(num_classes=_CIFAR_CLASSES),
        }),
        supervised_keys=('image', 'label'),
        homepage='https://github.com/hendrycks/robustness',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Return the test split of Cifar10.

    Args:
      dl_manager: download manager object.

    Returns:
      test split.
    """
    path = dl_manager.download_and_extract(_DOWNLOAD_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={'data_dir': os.path.join(path, _DIRNAME)},
        )
    ]

  def _generate_examples(self, data_dir):
    """Generate corrupted Cifar10 test data.

    Apply corruptions to the raw images according to self.corruption_type.

    Args:
      data_dir: root directory of downloaded dataset

    Yields:
      dictionary with image file and label.
    """
    corruption = self.builder_config.corruption
    severity = self.builder_config.severity

    images_file = os.path.join(data_dir, _CORRUPTIONS_TO_FILENAMES[corruption])
    labels_file = os.path.join(data_dir, _LABELS_FILENAME)

    with tf.io.gfile.GFile(labels_file, mode='rb') as f:
      labels = np.load(f)

    num_images = labels.shape[0] // 5
    # Labels are stacked 5 times so we can just read the first iteration
    labels = labels[:num_images]

    with tf.io.gfile.GFile(images_file, mode='rb') as f:
      images = np.load(f)

    # Slice images corresponding to correct severity level
    images = images[(severity - 1) * num_images : severity * num_images]

    for i, (image, label) in enumerate(zip(images, labels)):
      yield i, {
          'image': image,
          'label': label,
      }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to download remotely stored dataset source code."""

import dataclasses
import os
from typing import List

from etils import epath
from tensorflow_datasets.core import utils


@dataclasses.dataclass(frozen=True, eq=True)
class DatasetSource:
  """Structure containing information required to fetch a dataset.

  Attributes:
    root_path: Root directory of the dataset package (e.g.
      `gs://.../my_dataset/`, `github://.../my_dataset/`)
    filenames: Content of the dataset package
  """

  root_path: epath.Path
  filenames: List[str]

  @classmethod
  def from_json(cls, value: utils.JsonValue) -> 'DatasetSource':
    """Imports from JSON."""
    if isinstance(value, str):  # Single-file dataset ('.../my_dataset.py')
      path = epath.Path(value)
      return cls(root_path=path.parent, filenames=[path.name])
    elif isinstance(value, dict):  # Multi-file dataset
      return cls(
          root_path=epath.Path(value['root_path']),
          filenames=value['filenames'],
      )
    else:
      raise ValueError(f'Invalid input: {value}')

  def to_json(self) -> utils.JsonValue:
    """Exports to JSON."""
    if len(self.filenames) == 1:
      return os.fspath(self.root_path / self.filenames[0])
    else:
      return {
          'root_path': os.fspath(self.root_path),
          'filenames': self.filenames,
      }


def download_from_source(
    source: DatasetSource,
    dst: epath.Path,
) -> None:
  """Download the remote dataset code locally to the dst path.

  Args:
    source: Source of the dataset.
    dst: Empty directory to which the source will be copied
  """
  # Download the files
  # Note: We do not check for file existance before copying/downloading it to
  # not trigger unnecessary queries when `source.root_path` is `GithubPath`.
  # A `FileNotFoundError` is triggered if the file can't be downloaded.
  for filename in source.filenames:
    path = source.root_path / filename
    path.copy(dst / path.name)

  # Add the `__init__` file
  (dst / '__init__.py').write_text('')



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to process C4-WSRS."""

from __future__ import annotations
import collections
import dataclasses
import random
import re
from typing import Iterator, Mapping, Sequence, Tuple
import numpy as np

# Type annotations.
AbbrevExpansionPair = Tuple[str, str]


FREQUENT_EXPANSIONS = frozenset([
    'and',
    'by',
    'with',
    'home',
    'year',
    'years',
    'great',
    'two',
    'well',
    'work',
    'service',
    'company',
    'family',
    'quality',
    'day',
    'before',
    'every',
    'after',
    'small',
    'days',
    'old',
    'does',
    'please',
    'start',
    'center',
    'week',
    'high',
    'with a',
    'management',
    'food',
    'month',
    'north',
    'she',
    'black',
    'give',
    'county',
    'building',
    'right',
    'room',
    'four',
    'light',
    'call',
    'friday',
    'general',
    'companies',
    'times',
    'check',
    'west',
    'road',
    'report',
    'once',
    'heart',
    'five',
    'street',
    'point',
    'because',
    'size',
    'each',
    'east',
    'weeks',
    'united states',
    'type',
    'tuesday',
    'phone',
    'water',
    'december',
    'per',
    'second',
    'word',
    'established',
    'repair',
    'minutes',
    'regarding',
    'yesterday',
])


@dataclasses.dataclass
class WSRSFeatures:
  """Stores the original snippet and the snippet abbreviated by WSRS."""

  original_snippet: str = ''
  abbreviated_snippet: str = ''


def _get_backoff_prob(n: int, alpha: float) -> float:
  """Computes the back-off probability given current count and alpha value."""
  return (n + 1) ** -alpha


def create_word_finder_regex(words: Sequence[str]) -> re.Pattern[str]:
  """Creates regex pattern that matches any of the provided words.

  The regex pattern is comprised of 3 components: the prefix, the words, and the
  suffix. The words are a concatenation of all provided words separated by the
  OR operator. The prefix and suffix ensure that the matched word represents a
  distinct, delineated instance of that word.
    - prefix: A word must be preceded by any one of the following patterns:
      1. The string start
      2. Any character OTHER than alphanumeric or apostrophe
      3. An apostrophe preceded by a non-alphanumeric character (this
        ensures contractions and possessives do not lead to false positives)
    - suffix: A word must be followed by any one of the following patterns:
      1. The string end
      2. Any non-alphanumeric character
  Note: the order of the words matters, as the regex pattern will greedily
  match with the first word in the list. This has implications for instances
  in which a word is a distinguishable substring of another word.

  Args:
    words: A list of words that the pattern will find.

  Returns:
    A regex pattern that will match any instance of the provided words.
  """
  words_re = '|'.join([re.escape(word) for word in words])
  prefix_re = "^|[^\\w']|(?:^|\\W)'"
  suffix_re = '(?:$|\\W)'
  # We exclude the suffix from capture so that consecutive abbreviations
  # separated by spaces can be matched.
  re_string = f'(?:{prefix_re})({words_re})(?={suffix_re})'
  return re.compile(re_string)


def _get_abbreviation_expansion_pairs(
    snippet: str,
    abbreviations_by_expansion: Mapping[str, Sequence[str]],
    expansion_regex: re.Pattern[str],
) -> dict[int, AbbrevExpansionPair]:
  """Extracts possible abbreviation-expansion pairs from a snippet."""
  index_to_pair = {}
  for match in expansion_regex.finditer(snippet):
    match_start = match.span(1)[0]
    expansion = match.group(1)
    abbrev = random.choice(abbreviations_by_expansion[expansion])
    index_to_pair[match_start] = (abbrev, expansion)
  return index_to_pair


def extract_snippets(
    c4_doc: Mapping[str, bytes],
    max_sentences_per_snippet: int,
    abbreviations_by_expansion: Mapping[str, Sequence[str]],
    expansion_regex: re.Pattern[str],
    max_snippet_char_len: int,
    alpha_keep_no_rs: float,
    alpha_keep_rs: float,
) -> Iterator[
    tuple[
        str, tuple[str, AbbrevExpansionPair, Mapping[int, AbbrevExpansionPair]]
    ]
]:
  """Extracts variable-length multi-sentence snippets from C4 web text.

  Determines whether to keep or discard snippets based on the number of snippets
  previously kept which contain the same reverse substitution candidates (or
  lack thereof).

  Args:
    c4_doc: The C4 web crawl document.
    max_sentences_per_snippet: The maximum number of sentences that can be
      combined into a single snippet. The number of sentences combined into each
      snippet will be a randomly sampled integer from 1 to
      max_sentences_per_snippet.
    abbreviations_by_expansion: A mapping from expansion to all valid
      abbreviations for that expansion.
    expansion_regex: A regex pattern that will match any valid expansion.
    max_snippet_char_len: The maximum number of characters an original snippet
      can be composed of.
    alpha_keep_no_rs: Alpha value used to compute the probability of keeping a
      snippet containing no reverse substitutions. The complete formula is p =
      (n+1)**-alpha where n is the count of previously kept snippets from the
      same document.
    alpha_keep_rs: Alpha value used to compute the probability of keeping a
      snippet containing a particular abbreviation-expansion reverse
      substitution candidate. The complete formula is p = (n+1)**-alpha where n
      is the count of previously kept snippets from the same document containing
      the same reverse substitution candidate.

  Yields:
    A key-value tuple where the key is a string containing the document url and
    incremental snippet id, and the value is a 3-tuple containing the snippet,
    the rarest abbreviation-expansion substitution candidate in the snippet, and
    a mapping from snippet index to each substitution candidate.
  """
  count_by_pair = collections.Counter()
  text = c4_doc['text'].decode()
  text = text.replace('\n', ' ')
  sentences = text.split('. ')
  sentence_idx = 0
  snippet_id = 0
  while sentence_idx < len(sentences):
    num_sentences_sampled = np.random.randint(1, max_sentences_per_snippet + 1)
    snippet = '. '.join(
        sentences[sentence_idx : sentence_idx + num_sentences_sampled]
    )
    # Period added stochastically to make models robust to the absence of
    # ending punctuation.
    if not snippet.endswith('.') and np.random.uniform() < 0.5:
      snippet += '.'
    key = f'url={c4_doc["url"].decode()},snippet_id={snippet_id}'
    snippet_id += 1
    sentence_idx += num_sentences_sampled
    snippet = snippet.lower().strip()
    if len(snippet) > max_snippet_char_len:
      continue
    index_to_pair = _get_abbreviation_expansion_pairs(
        snippet, abbreviations_by_expansion, expansion_regex
    )
    if index_to_pair:
      smallest_count, rarest_pair = None, None
      for pair in index_to_pair.values():
        if pair[1] in FREQUENT_EXPANSIONS:
          continue
        count = count_by_pair[pair]
        if smallest_count is None or count < smallest_count:
          smallest_count = count
          rarest_pair = pair
      sample_prob = (
          _get_backoff_prob(smallest_count, alpha=alpha_keep_rs)
          if smallest_count is not None
          else 0.0
      )
      if np.random.uniform() >= sample_prob:
        continue
      # Count should only be increased if snippet is sampled.
      for pair in index_to_pair.values():
        count_by_pair[pair] += 1
    else:
      rarest_pair = ('', '')
      index_to_pair = {0: ('', '')}
      count = count_by_pair[('', '')]
      sample_prob = _get_backoff_prob(count + 100, alpha=alpha_keep_no_rs)
      if np.random.uniform() >= sample_prob:
        continue
      count_by_pair[('', '')] += 1

    yield key, (snippet, rarest_pair, index_to_pair)  # pytype: disable=bad-return-type  # container-simplification


def _reverse_substitute_snippet(
    snippet: str,
    index_to_pair: Mapping[int, AbbrevExpansionPair],
    substitution_rate: float,
) -> str:
  """Reverse substitutes a snippet, swapping expansions for abbreviations.

  For each candidate abbreviation, its abbreviation is sampled using the
  substitution_rate value.

  Args:
    snippet: The snippet to abbreviate.
    index_to_pair: A mapping from snippet index to reverse-substitution
      candidate for that index.
    substitution_rate: The rate at which expansions should be substituted for
      abbreviations.

  Returns:
    The abbreviated snippet.
  """
  abbreviated_snippet = ''
  for index, (abbrev, expansion) in sorted(index_to_pair.items(), reverse=True):
    if np.random.uniform() >= substitution_rate:
      continue
    abbreviated_snippet = (
        abbrev + snippet[index + len(expansion) :] + abbreviated_snippet
    )
    snippet = snippet[:index]
  abbreviated_snippet = snippet + abbreviated_snippet
  return abbreviated_snippet


def reverse_substitution(
    element: tuple[
        str, tuple[str, AbbrevExpansionPair, Mapping[int, AbbrevExpansionPair]]
    ],
    substitution_rate: float,
    min_snippet_token_len: int,
) -> Iterator[tuple[AbbrevExpansionPair, tuple[str, WSRSFeatures]]]:
  """Conducts reverse substitution.

  Yields one WSRSFeatures obj per unique replacement.

  Args:
    element: A key-value tuple containing the url as the key, and a 3-tuple as
      the value containing the snippet extracted from the c4 doc, the rarest
      abbreviation-expansion substitution candidate in the snippet, and a
      mapping from snippet index to each substitution candidate.
    substitution_rate: The rate at which expansions should be substituted for
      abbreviations.
    min_snippet_token_len: The minimum number of tokens an abbreviated snippet
      can be composed of.

  Yields:
    A key-value tuple in which the key is a tuple containing the rarest
    abbreviation-expansion pair replacement for that snippet, and the value is a
    tuple containing the original key and a WSRSFeatures object holding the
    extracted snippet and its abbreviated form.
  """
  key, (snippet, rarest_pair, index_to_pair) = element
  abbreviated_snippet = snippet
  if index_to_pair:
    abbreviated_snippet = _reverse_substitute_snippet(
        snippet, index_to_pair, substitution_rate
    )
  if len(abbreviated_snippet.split(' ')) < min_snippet_token_len:
    return
  features = WSRSFeatures()
  features.original_snippet = snippet
  features.abbreviated_snippet = abbreviated_snippet

  yield rarest_pair, (key, features)


def sample_snippets_by_substitution(
    snippets_by_substitution: tuple[
        AbbrevExpansionPair, Sequence[tuple[str, WSRSFeatures]]
    ],
    num_snippets_per_substitution: int,
) -> Iterator[tuple[str, WSRSFeatures]]:
  """Samples a fixed number of snippets for each substitution pair."""
  limit = num_snippets_per_substitution
  if limit == 0:
    return
  snippets = snippets_by_substitution[1]
  for key, features in snippets:
    yield key, features
    limit -= 1
    if limit <= 0:
      break



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test case util to test `tfds.features.FeatureConnector`."""

import contextlib
import dataclasses
import functools
from typing import Any, Optional, Type

import dill
from etils import enp
import numpy as np
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import features
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.testing import test_case
from tensorflow_datasets.testing import test_utils


@dataclasses.dataclass
class FeatureExpectationItem:
  """Test item of a FeatureExpectation.

  Should be passed to `assertFeature` method of `FeatureExpectationsTestCase`.

  Each `FeatureExpectationItem` test an example serialization/deserialization (
  `feature.encode_example` -> `example_serializer.serialize_example` ->
  `example_parse.parse_example` -> `feature.decode_example`).

  Attributes:
    value: Input to `features.encode_example`
    expected: Expected output after `features.decode_example`
    expected_np: Expected output after `features.decode_example_np`
    expected_serialized: Optional
    decoders: Optional `tfds.decode.Decoder` objects (to overwrite the default
      `features.decode_example`). See
      https://www.tensorflow.org/datasets/decode.
    dtype: If `decoders` is provided, the output of `decode_example` is checked
      against this value (otherwise, output is checked against `features.dtype`)
    shape: If `decoders` is provided, the output of `decode_example` is checked
      against this value (otherwise, output is checked against `features.shape`)
    raise_cls: Expected error raised during `features.encode_example`. When set
      `expected` and `expected_serialized` should be `None`.
    raise_cls_np: Expected error raised during `features.encode_example_np`.
      When set `expected_np` and `expected_serialized_np` should be `None`.
    raise_msg: Expected error message regex.
    atol: If provided, compare float values with this precision (use default
      otherwise).
  """

  value: Any
  expected: Optional[Any] = None
  expected_np: Optional[np.ndarray] = None
  expected_serialized: Optional[Any] = None
  decoders: Optional[utils.TreeDict[Any]] = None
  dtype: Optional[tf.dtypes.DType] = None
  shape: Optional[utils.Shape] = None
  raise_cls: Optional[Type[Exception]] = None
  raise_cls_np: Optional[Type[Exception]] = None
  raise_msg: Optional[str] = None
  atol: Optional[float] = None

  def __post_init__(self):
    if not self.decoders and (self.dtype is not None or self.shape is not None):
      raise ValueError('dtype and shape should only be set with transform')


class SubTestCase(test_case.TestCase):
  """Adds subTest() context manager to the TestCase if supported.

  Note: To use this feature, make sure you call super() in setUpClass to
  initialize the sub stack.
  """

  @classmethod
  def setUpClass(cls):
    super(SubTestCase, cls).setUpClass()
    cls._sub_test_stack = []

  @contextlib.contextmanager
  def _subTest(self, test_str):
    self._sub_test_stack.append(test_str)
    sub_test_str = '/'.join(self._sub_test_stack)
    with self.subTest(sub_test_str):
      yield
    self._sub_test_stack.pop()

  def assertAllEqualNested(self, d1, d2, *, atol: Optional[float] = None):
    """Same as assertAllEqual but compatible with nested dict.

    Args:
      d1: First element to compare
      d2: Second element to compare
      atol: If given, perform a close float comparison. Otherwise, perform an
        exact comparison
    """
    if isinstance(d1, dict):
      # assertAllEqual do not works well with dictionaries so assert
      # on each individual elements instead
      zipped_examples = utils.zip_nested(d1, d2, dict_only=True)
      utils.map_nested(
          # recursively call assertAllEqualNested in case there is a dataset.
          lambda x: self.assertAllEqualNested(x[0], x[1], atol=atol),
          zipped_examples,
          dict_only=True,
      )
    elif isinstance(d1, (tf.data.Dataset, dataset_utils._IterableDataset)):  # pylint: disable=protected-access
      # Checks length and elements of the dataset. At the moment, more than one
      # level of nested datasets is not supported.
      self.assertEqual(len(d1), len(d2))
      for ex1, ex2 in zip(d1, d2):
        self.assertAllEqualNested(ex1, ex2, atol=atol)
    elif atol:
      self.assertAllClose(d1, d2, atol=atol)
    else:
      self.assertAllEqual(d1, d2)


class RaggedConstant(object):
  """Container of tf.ragged.constant values.

  This simple wrapper forward the arguments to delay the RaggedTensor
  construction after `@run_in_graph_and_eager_modes` has been called.
  This is required to avoid incompabilities between Graph/eager.
  """

  def __init__(self, *args, **kwargs):
    self._args = args
    self._kwargs = dict(kwargs)

  def build(self):
    return tf.ragged.constant(*self._args, **self._kwargs)


class FeatureExpectationsTestCase(SubTestCase):
  """Tests FeatureExpectations with full encode-decode."""

  @test_utils.run_in_graph_and_eager_modes()
  def assertFeature(
      self,
      feature,
      shape,
      dtype,
      tests,
      serialized_info=None,
      # TODO(b/227584124): remove this parameter after fixing this bug
      test_tensor_spec=True,
      skip_feature_tests=False,
      test_attributes=None,
  ):
    """Test the given feature against the predicates."""
    self.assertFeatureEagerOnly(
        feature=feature,
        shape=shape,
        dtype=dtype,
        tests=tests,
        serialized_info=serialized_info,
        test_tensor_spec=test_tensor_spec,
        skip_feature_tests=skip_feature_tests,
        test_attributes=test_attributes,
    )

  def assertFeatureEagerOnly(
      self,
      feature,
      shape,
      dtype,
      tests,
      serialized_info=None,
      test_tensor_spec=True,
      skip_feature_tests=False,
      test_attributes=None,
  ):
    """Test the given feature against the predicates."""
    # Fill kwargs
    run_tests = functools.partial(
        self._run_tests,
        shape=shape,
        dtype=dtype,
        tests=tests,
        test_tensor_spec=test_tensor_spec,
    )
    assert_feature = functools.partial(
        self._assert_feature,
        shape=shape,
        dtype=dtype,
        tests=tests,
        serialized_info=serialized_info,
        test_tensor_spec=test_tensor_spec,
        skip_feature_tests=skip_feature_tests,
        test_attributes=test_attributes,
    )

    # Create the feature dict
    fdict = features.FeaturesDict({'inner': feature})

    # Check whether the following doesn't raise an exception
    fdict.catalog_documentation()

    with self._subTest('feature'):
      assert_feature(feature=feature)
      run_tests(serialize_fdict=fdict, deserialize_fdict=fdict, feature=feature)

    # Test the feature again to make sure it behave correctly after restoring
    # TODO(tfds): Remove `skip_feature_tests` after text encoders are removed
    if not skip_feature_tests:
      # Restored from config
      with test_utils.tmp_dir() as config_dir:
        feature.save_config(config_dir)
        new_feature = feature.from_config(config_dir)
      assert_feature(feature=new_feature)
      with self._subTest('feature_roundtrip'):
        run_tests(
            serialize_fdict=fdict, deserialize_fdict=fdict, feature=new_feature
        )

      # Restored from proto
      with test_utils.tmp_dir() as config_dir:
        feature_proto = feature.to_proto()
        feature.save_metadata(config_dir, feature_name=None)
        new_feature = feature_lib.FeatureConnector.from_proto(feature_proto)
        new_feature.load_metadata(config_dir, feature_name=None)
      assert_feature(feature=new_feature)
      new_fdict = features.FeaturesDict({'inner': new_feature})
      with self._subTest('feature_proto_roundtrip'):
        run_tests(
            serialize_fdict=fdict, deserialize_fdict=fdict, feature=new_feature
        )
      with self._subTest('serialize_fdict'):
        run_tests(
            serialize_fdict=new_fdict,
            deserialize_fdict=fdict,
            feature=new_feature,
        )
      with self._subTest('deserialize_fdict'):
        run_tests(
            serialize_fdict=fdict,
            deserialize_fdict=new_fdict,
            feature=new_feature,
        )

  def _assert_feature(
      self,
      feature,
      shape,
      dtype,
      tests,
      serialized_info=None,
      test_tensor_spec=True,
      skip_feature_tests=False,
      test_attributes=None,
  ):
    with self._subTest('shape'):
      self.assertEqual(feature.shape, shape)
    with self._subTest('dtype'):
      self.assertEqual(feature.dtype, dtype)
      tree_utils.map_structure(enp.lazy.is_np_dtype, feature.np_dtype)
      tree_utils.map_structure(enp.lazy.is_tf_dtype, feature.tf_dtype)

    # Check the serialized features
    if serialized_info:
      with self._subTest('serialized_info'):
        self.assertEqual(
            serialized_info,
            feature.get_serialized_info(),
        )

    if not skip_feature_tests and test_attributes:
      for key, value in test_attributes.items():
        self.assertEqual(getattr(feature, key), value)

  def _run_tests(
      self,
      serialize_fdict,
      deserialize_fdict,
      feature,
      shape,
      dtype,
      tests,
      test_tensor_spec=True,
  ):
    for i, test in enumerate(tests):
      with self._subTest(str(i)):
        self.assertFeatureTest(
            serialize_fdict=serialize_fdict,
            deserialize_fdict=deserialize_fdict,
            test=test,
            feature=feature,
            shape=shape,
            dtype=dtype,
            test_tensor_spec=test_tensor_spec,
        )

  def assertFeatureTest(
      self,
      serialize_fdict,
      deserialize_fdict,
      test,
      feature,
      shape,
      dtype,
      test_tensor_spec: bool = True,
  ):
    """Test that encode=>decoding of a value works correctly."""
    # test feature.encode_example can be pickled and unpickled for beam.
    dill.loads(dill.dumps(feature.encode_example))

    input_value = {'inner': test.value}

    if test.raise_cls is not None or test.raise_cls_np is not None:
      if test.raise_cls is not None:
        with self._subTest('raise'):
          if not test.raise_msg:
            raise ValueError(
                'test.raise_msg should be set with {} for test {}'.format(
                    test.raise_cls, type(feature)
                )
            )
          with self.assertRaisesWithPredicateMatch(
              test.raise_cls, test.raise_msg
          ):
            features_encode_decode(
                serialize_fdict,
                deserialize_fdict,
                input_value,
                decoders=test.decoders,
            )
      if test.raise_cls_np is not None:
        with self._subTest('raise_np'):
          if not test.raise_msg:
            raise ValueError(
                'test.raise_msg should be set with {} for test {}'.format(
                    test.raise_cls_np, type(feature)
                )
            )
          with self.assertRaisesWithPredicateMatch(
              test.raise_cls_np, test.raise_msg
          ):
            features_encode_decode_np(
                serialize_fdict,
                deserialize_fdict,
                input_value,
                decoders=test.decoders,
            )
    else:
      # Test the serialization only
      if test.expected_serialized is not None:
        with self._subTest('out_serialize'):
          self.assertEqual(
              test.expected_serialized,
              feature.encode_example(test.value),
          )

      # Test serialization + decoding from disk for NumPy worflow
      if test.expected_np is not None:
        with self._subTest('out_np'):
          out_numpy = features_encode_decode_np(
              serialize_fdict,
              deserialize_fdict,
              input_value,
              decoders={'inner': test.decoders},
          )
          with self._subTest('out_np_value'):
            np.testing.assert_array_equal(out_numpy['inner'], test.expected_np)

      # Test serialization + decoding from disk
      with self._subTest('out'):
        out_tensor, out_numpy, out_element_spec = features_encode_decode(
            serialize_fdict,
            deserialize_fdict,
            input_value,
            decoders={'inner': test.decoders},
        )
        out_tensor = out_tensor['inner']
        out_numpy = out_numpy['inner']
        out_element_spec = out_element_spec['inner']

        if test_tensor_spec:
          with self._subTest('tensor_spec'):
            assert feature.get_tensor_spec() == out_element_spec

        # Assert the returned type match the expected one
        with self._subTest('dtype'):

          def _get_dtype(s):
            if isinstance(s, tf.data.Dataset):
              return tf.nest.map_structure(_get_dtype, s.element_spec)
            else:
              return s.dtype

          out_dtypes = tf.nest.map_structure(_get_dtype, out_tensor)
          self.assertEqual(out_dtypes, test.dtype or feature.dtype)
        with self._subTest('shape'):
          # For shape, because (None, 3) match with (5, 3), we use
          # tf.TensorShape.assert_is_compatible_with on each of the elements
          expected_shape = feature.shape if test.shape is None else test.shape

          def _get_shape(s):
            if isinstance(s, tf.data.Dataset):
              return utils.map_nested(_get_shape, s.element_spec)
            else:
              return s.shape

          out_shapes = utils.map_nested(_get_shape, out_tensor)

          shapes_tuple = utils.zip_nested(out_shapes, expected_shape)
          utils.map_nested(
              lambda x: x[0].assert_is_compatible_with(x[1]), shapes_tuple
          )

        # Assert value
        with self._subTest('out_value'):
          # Eventually construct the tf.RaggedTensor
          expected = tf.nest.map_structure(
              lambda t: t.build() if isinstance(t, RaggedConstant) else t,
              test.expected,
          )
          self.assertAllEqualNested(out_numpy, expected, atol=test.atol)

        # Assert the HTML representation works
        if not test.decoders:
          with self._subTest('repr'):
            self._test_repr(feature, out_numpy)

  def _test_repr(
      self,
      feature: features.FeatureConnector,
      out_numpy: np.ndarray,
  ) -> None:
    """Test that the HTML repr works."""
    # pylint: disable=protected-access
    flat_example = feature._flatten(out_numpy)
    flat_features = feature._flatten(feature)
    flat_serialized_info = feature._flatten(feature.get_serialized_info())
    # pylint: enable=protected-access
    for ex, f, spec in zip(flat_example, flat_features, flat_serialized_info):
      # Features with multi-data not supported
      if isinstance(spec, dict):
        continue
      # TODO(tfds): Should use `as_dataframe._get_feature` instead, to
      # correctly compute for `sequence_rank` for subclasses like `Video`.
      elif spec.sequence_rank == 0:
        text = f.repr_html(ex)
      elif spec.sequence_rank == 1:
        text = f.repr_html_batch(ex)
      elif spec.sequence_rank > 1:
        text = f.repr_html_ragged(ex)
      self.assertIsInstance(text, str)


def features_encode_decode(
    serialize_fdict, deserialize_fdict, example, decoders
):
  """Runs the full pipeline: encode > write > tmp files > read > decode."""
  # Serialize/deserialize the example using TensorFlow methods.
  serialized_example = serialize_fdict.serialize_example(example)

  decode_fn = functools.partial(
      deserialize_fdict.deserialize_example,
      decoders=decoders,
  )
  ds = tf.data.Dataset.from_tensors(serialized_example)
  ds = ds.map(decode_fn)

  if tf.executing_eagerly():
    out_tensor = next(iter(ds))
  else:
    out_tensor = tf.compat.v1.data.make_one_shot_iterator(ds).get_next()
  out_numpy = dataset_utils.as_numpy(out_tensor)
  return out_tensor, out_numpy, ds.element_spec


def features_encode_decode_np(
    serialize_fdict, deserialize_fdict, example, decoders
):
  """Runs the full pipeline: encode > write > tmp files > read > decode."""
  # Serialize/deserialize the example using NumPy methods.
  serialized_example = serialize_fdict.serialize_example(example)
  deserialized_example = deserialize_fdict.deserialize_example_np(
      serialized_example, decoders=decoders
  )
  return deserialized_example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wikipedia-based Image Text (WIT) Dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.vision_language.wit import wit


class WitTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for wit dataset."""

  DATASET_CLASS = wit.Wit
  SPLITS = {
      'train': 3,
      'val': 1,
      'test': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': ['train-00000-of-00009.tsv'],
      'val': ['val-00000-of-00005.tsv'],
      'test': ['test-00000-of-00005.tsv'],
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adapters for file formats."""

from __future__ import annotations

import abc
import enum
import os
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Type, Union

from etils import epath
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import array_record_module
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

ExamplePositions = List[Any]


class FileFormat(enum.Enum):
  """Format of the record files.

  The values of the enumeration are used as filename endings/suffix.
  """

  TFRECORD = 'tfrecord'
  RIEGELI = 'riegeli'
  ARRAY_RECORD = 'array_record'

  @property
  def file_suffix(self) -> str:
    return ADAPTER_FOR_FORMAT[self].FILE_SUFFIX

  @classmethod
  def from_value(cls, file_format: Union[str, 'FileFormat']) -> 'FileFormat':
    try:
      return cls(file_format)
    except ValueError as e:
      all_values = [f.value for f in cls]
      raise ValueError(
          f'{file_format} is not a valid FileFormat! '
          f'Valid file formats: {all_values}'
      ) from e


DEFAULT_FILE_FORMAT = FileFormat.TFRECORD


class FileAdapter(abc.ABC):
  """Interface for Adapter objects which read and write examples in a format."""

  FILE_SUFFIX: ClassVar[str]
  BUFFER_SIZE = 8 << 20  # 8 MiB per file.

  @classmethod
  @abc.abstractmethod
  def make_tf_data(
      cls,
      filename: epath.PathLike,
      buffer_size: Optional[int] = None,
  ) -> tf.data.Dataset:
    """Returns TensorFlow Dataset comprising given record file."""
    raise NotImplementedError()

  @classmethod
  @abc.abstractmethod
  def write_examples(
      cls,
      path: epath.PathLike,
      iterator: Iterable[type_utils.KeySerializedExample],
  ) -> Optional[ExamplePositions]:
    """Write examples from given iterator in given path.

    Args:
      path: Path where to write the examples.
      iterator: Iterable of examples.

    Returns:
      List of record positions for each record in the given iterator. In case of
      TFRecords, does not return anything.
    """
    raise NotImplementedError()


class TfRecordFileAdapter(FileAdapter):
  """File adapter for TFRecord file format."""

  FILE_SUFFIX = 'tfrecord'

  @classmethod
  def make_tf_data(
      cls,
      filename: epath.PathLike,
      buffer_size: Optional[int] = None,
  ) -> tf.data.Dataset:
    """Returns TensorFlow Dataset comprising given record file."""
    buffer_size = buffer_size or cls.BUFFER_SIZE
    return tf.data.TFRecordDataset(filename, buffer_size=buffer_size)

  @classmethod
  def write_examples(
      cls,
      path: epath.PathLike,
      iterator: Iterable[type_utils.KeySerializedExample],
  ) -> Optional[ExamplePositions]:
    """Write examples from given iterator in given path.

    Args:
      path: Path where to write the examples.
      iterator: Iterable of examples.

    Returns:
      None
    """
    with tf.io.TFRecordWriter(os.fspath(path)) as writer:
      for _, serialized_example in iterator:
        writer.write(serialized_example)
      writer.flush()


class RiegeliFileAdapter(FileAdapter):
  """File adapter for Riegeli file format."""

  FILE_SUFFIX = 'riegeli'

  @classmethod
  def make_tf_data(
      cls,
      filename: epath.PathLike,
      buffer_size: Optional[int] = None,
  ) -> tf.data.Dataset:
    buffer_size = buffer_size or cls.BUFFER_SIZE
    from riegeli.tensorflow.ops import riegeli_dataset_ops as riegeli_tf  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error

    return riegeli_tf.RiegeliDataset(filename, buffer_size=buffer_size)

  @classmethod
  def write_examples(
      cls,
      path: epath.PathLike,
      iterator: Iterable[type_utils.KeySerializedExample],
  ) -> Optional[ExamplePositions]:
    """Write examples from given iterator in given path.

    Args:
      path: Path where to write the examples.
      iterator: Iterable of examples.

    Returns:
      List of record positions for each record in the given iterator.
    """
    positions = []
    import riegeli  # pylint: disable=g-import-not-at-top

    with tf.io.gfile.GFile(os.fspath(path), 'wb') as f:
      with riegeli.RecordWriter(f, options='transpose') as writer:
        for _, record in iterator:
          writer.write_record(record)
          positions.append(writer.last_pos)
    return positions


class ArrayRecordFileAdapter(FileAdapter):
  """File adapter for ArrayRecord file format."""

  FILE_SUFFIX = 'array_record'

  @classmethod
  def make_tf_data(
      cls,
      filename: epath.PathLike,
      buffer_size: Optional[int] = None,
  ) -> tf.data.Dataset:
    """Returns TensorFlow Dataset comprising given array record file."""
    raise NotImplementedError(
        '`.as_dataset()` not implemented for ArrayRecord files. Please, use'
        ' `.as_data_source()`.'
    )

  @classmethod
  def write_examples(
      cls,
      path: epath.PathLike,
      iterator: Iterable[type_utils.KeySerializedExample],
  ) -> Optional[ExamplePositions]:
    """Write examples from given iterator in given path.

    Args:
      path: Path where to write the examples.
      iterator: Iterable of examples.

    Returns:
      None
    """
    writer = array_record_module.ArrayRecordWriter(
        os.fspath(path), 'group_size:1'
    )
    for _, serialized_example in iterator:
      writer.write(serialized_example)
    writer.close()


def _to_bytes(key: type_utils.Key) -> bytes:
  """Convert the key to bytes."""
  if isinstance(key, int):
    return key.to_bytes(128, byteorder='big')  # Use 128 as this match md5
  elif isinstance(key, bytes):
    return key
  elif isinstance(key, str):
    return key.encode('utf-8')
  else:
    raise TypeError(f'Invalid key type: {type(key)}')


# Create a mapping from FileFormat -> FileAdapter.
ADAPTER_FOR_FORMAT: Dict[FileFormat, Type[FileAdapter]] = {
    FileFormat.RIEGELI: RiegeliFileAdapter,
    FileFormat.TFRECORD: TfRecordFileAdapter,
    FileFormat.ARRAY_RECORD: ArrayRecordFileAdapter,
}

_FILE_SUFFIX_TO_FORMAT = {
    adapter.FILE_SUFFIX: file_format
    for file_format, adapter in ADAPTER_FOR_FORMAT.items()
}


def file_format_from_suffix(file_suffix: str) -> FileFormat:
  """Returns the file format associated with the file extension (`tfrecord`)."""
  if file_suffix not in _FILE_SUFFIX_TO_FORMAT:
    raise ValueError(
        'Unrecognized file extension: Should be one of '
        f'{list(_FILE_SUFFIX_TO_FORMAT.values())}'
    )
  return _FILE_SUFFIX_TO_FORMAT[file_suffix]


def is_example_file(filename: str) -> bool:
  """Whether the given filename is a record file."""
  return any(
      f'.{adapter.FILE_SUFFIX}' in filename
      for adapter in ADAPTER_FOR_FORMAT.values()
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the reader config."""

from __future__ import annotations

import dataclasses
import enum
from typing import Callable, Optional, Sequence, Union, cast

from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

InterleaveSortFn = Callable[
    [Sequence[shard_utils.FileInstruction]],
    Sequence[shard_utils.FileInstruction],
]


class _MISSING(str):
  pass


MISSING = _MISSING('missing')


@dataclasses.dataclass(eq=False)
class ReadConfig:
  # pyformat: disable
  """Configures input reading pipeline.

  Attributes:
    options: `tf.data.Options()`, dataset options to use. Note that when
      `shuffle_files` is True and no seed is defined, deterministic will be set
      to False internally, unless it is defined here.
    try_autocache: If True (default) and the dataset satisfy the right
      conditions (dataset small enough, files not shuffled,...) the dataset will
      be cached during the first iteration (through `ds = ds.cache()`).
    repeat_filenames: If True, repeat the filenames iterator. This will result
      in an infinite dataset. Repeat is called after the shuffle of the
      filenames.
    add_tfds_id: If True, examples `dict` in `tf.data.Dataset` will have an
      additional key `'tfds_id': tf.Tensor(shape=(), dtype=tf.string)`
      containing the example unique identifier (e.g.
      'train.tfrecord-000045-of-001024__123').
       Note: IDs might changes in future version of TFDS.
    shuffle_seed: `tf.int64`, seed forwarded to `tf.data.Dataset.shuffle` during
      file shuffling (which happens when `tfds.load(..., shuffle_files=True)`).
    shuffle_reshuffle_each_iteration: `bool`, forwarded to
      `tf.data.Dataset.shuffle` during file shuffling (which happens when
      `tfds.load(..., shuffle_files=True)`).
    interleave_cycle_length: `int`, forwarded to `tf.data.Dataset.interleave`.
    interleave_block_length: `int`, forwarded to `tf.data.Dataset.interleave`.
    input_context: `tf.distribute.InputContext`, if set, each worker will read a
      different set of file. For more info, see the
      [distribute_datasets_from_function
      documentation](https://www.tensorflow.org/api_docs/python/tf/distribute/Strategy#experimental_distribute_datasets_from_function).
      Note:  * Each workers will always read the same subset of files.
        `shuffle_files` only shuffle files within each worker. * If
        `info.splits[split].num_shards < input_context.num_input_pipelines`, an
        error will be raised, as some workers would be empty.
    experimental_interleave_sort_fn: Function with signature `List[FileDict] ->
      List[FileDict]`, which takes the list of `dict(file: str, take: int, skip:
      int)` and returns the modified version to read. This can be used to
      sort/shuffle the shards to read in a custom order, instead of relying on
      `shuffle_files=True`.
    skip_prefetch: If False (default), add a `ds.prefetch()` op at the end.
      Might be set for performance optimization in some cases (e.g. if you're
      already calling `ds.prefetch()` at the end of your pipeline)
    num_parallel_calls_for_decode: The number of parallel calls for decoding
      record. By default using tf.data's AUTOTUNE.
    num_parallel_calls_for_interleave_files: The number of parallel calls for
      interleaving files. By default using tf.data's AUTOTUNE.
    enable_ordering_guard: When True (default), an exception is raised if
      shuffling or interleaving are used on an ordered dataset.
    assert_cardinality: When True (default), an exception is raised if at the
      end of an Epoch the number of read examples does not match the expected
      number from dataset metadata. A power user would typically want to set
      False if input files have been tempered with and they don't mind missing
      records or have too many of them.
    override_buffer_size: number of bytes to pass to file readers for buffering.
  """
  # pyformat: enable

  # General tf.data.Dataset parametters
  options: Optional[tf.data.Options] = None
  try_autocache: bool = True
  repeat_filenames: bool = False
  add_tfds_id: bool = False
  # tf.data.Dataset.shuffle parameters
  shuffle_seed: Optional[int] = None
  shuffle_reshuffle_each_iteration: Optional[bool] = None
  # Interleave parameters
  # Ideally, we should switch interleave values to None to dynamically set
  # those value depending on the user system. However, this would make the
  # generation order non-deterministic accross machines.
  interleave_cycle_length: Union[Optional[int], _MISSING] = MISSING
  interleave_block_length: Optional[int] = 16
  input_context: Optional[tf.distribute.InputContext] = None
  experimental_interleave_sort_fn: Optional[InterleaveSortFn] = None
  skip_prefetch: bool = False
  num_parallel_calls_for_decode: Optional[int] = None
  # Cast to an `int`. `__post_init__` will ensure the type invariant.
  num_parallel_calls_for_interleave_files: Optional[int] = cast(int, MISSING)
  enable_ordering_guard: bool = True
  assert_cardinality: bool = True
  override_buffer_size: Optional[int] = None

  def __post_init__(self):
    self.options = self.options or tf.data.Options()
    if self.num_parallel_calls_for_decode is None:
      self.num_parallel_calls_for_decode = tf.data.AUTOTUNE
    if self.num_parallel_calls_for_interleave_files == MISSING:
      self.num_parallel_calls_for_interleave_files = tf.data.AUTOTUNE



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""web_nlg dataset."""

from __future__ import annotations

import os
import xml.etree.ElementTree as etree

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@inproceedings{gardent2017creating,
    title = ""Creating Training Corpora for {NLG} Micro-Planners"",
    author = ""Gardent, Claire  and
      Shimorina, Anastasia  and
      Narayan, Shashi  and
      Perez-Beltrachini, Laura"",
    booktitle = ""Proceedings of the 55th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)"",
    month = jul,
    year = ""2017"",
    address = ""Vancouver, Canada"",
    publisher = ""Association for Computational Linguistics"",
    doi = ""10.18653/v1/P17-1017"",
    pages = ""179--188"",
    url = ""https://www.aclweb.org/anthology/P17-1017.pdf""
}
"""

_DESCRIPTION = """
The data contains sets of 1 to 7 triples of the form subject-predicate-object
extracted from (DBpedia)[https://wiki.dbpedia.org/] and natural language text
that's a verbalisation of these triples.
The test data spans 15 different domains where only 10 appear in the training
data.
The dataset follows a standarized table format.
"""

_URL = 'https://drive.google.com/uc?export=download&id=1C3d0a1wPkJqI3SVyYNtl99J1lErYYhsc'


class UnexpectedFormatError(Exception):
  """The specification of the sample set is malformed."""


class WebNlg(tfds.core.GeneratorBasedBuilder):
  """Set of triples subject-predicate-object to text."""

  VERSION = tfds.core.Version('0.1.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        # This is the description that will appear on the datasets page.
        description=_DESCRIPTION,
        # tfds.features.FeatureConnectors
        features=tfds.features.FeaturesDict({
            'input_text': {
                'table': tfds.features.Sequence({
                    # Each row will be one triple fact.
                    # we'll only have subject/predicate/object headers
                    'column_header': np.str_,
                    'row_number': np.int16,
                    'content': np.str_,
                }),
                # context will be the category
                'context': np.str_,
            },
            'target_text': np.str_,
        }),
        supervised_keys=('input_text', 'target_text'),
        # Homepage of the dataset for documentation
        homepage='https://webnlg-challenge.loria.fr/challenge_2017/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    def get_files_in_dir(directory):
      all_files = []
      for path, _, files in tf.io.gfile.walk(directory):
        for name in files:
          all_files.append(os.path.join(path, name))
      return all_files

    extracted_path = os.path.join(
        dl_manager.download_and_extract(_URL),
        'webnlg-dataset-master',
        'webnlg_challenge_2017',
    )
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'list_files': get_files_in_dir(
                    os.path.join(extracted_path, 'train')
                ),
                'set_name': 'train',
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                'list_files': get_files_in_dir(
                    os.path.join(extracted_path, 'dev')
                ),
                'set_name': 'validation',
            },
        ),
        tfds.core.SplitGenerator(
            name='test_unseen',
            gen_kwargs={
                'list_files': [
                    os.path.join(
                        extracted_path, 'test', 'testdata_unseen_with_lex.xml'
                    )
                ],
                'set_name': 'test_unseen',
            },
        ),
        tfds.core.SplitGenerator(
            name='test_all',
            gen_kwargs={
                'list_files': [
                    os.path.join(
                        extracted_path, 'test', 'testdata_with_lex.xml'
                    )
                ],
                'set_name': 'test_all',
            },
        ),
    ]

  def _generate_examples(self, list_files, set_name):
    """Yields examples."""
    for file_path in list_files:
      with tf.io.gfile.GFile(file_path, 'r') as f:
        xml_file = etree.parse(f)
        xml_root = xml_file.getroot()
        for entry in list(xml_root)[0]:
          category = entry.attrib['category']
          entry_id = '{}_{}'.format(
              os.path.basename(file_path), entry.attrib['eid']
          )
          triples_set = []
          target_text_i = 0
          for child_element in entry:
            if child_element.tag == 'modifiedtripleset':
              for i, triple in enumerate(child_element):
                text = triple.text
                if not text:
                  continue
                for header, content in zip(
                    ['subject', 'predicate', 'object'], text.split(' | ')
                ):
                  triples_set.append({
                      'column_header': header,
                      'row_number': i,
                      'content': content,
                  })
            elif child_element.tag == 'lex':
              if not triples_set:
                raise UnexpectedFormatError(
                    'Found language expresion with no previous triplesets.'
                )
              yield '{}_#{}'.format(entry_id, target_text_i), {
                  'input_text': {'table': triples_set, 'context': category},
                  'target_text': child_element.text,
              }
              target_text_i += 1



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""D4RL Ant-maze dataset."""

from typing import Any

from tensorflow_datasets.d4rl import dataset_builder
import tensorflow_datasets.public_api as tfds


class D4rlAntmaze(dataset_builder.D4RLDatasetBuilder):
  """DatasetBuilder for ant dataset."""

  VERSION = tfds.core.Version('1.1.1')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.1.1': 'Added v2 datasets.',
  }

  _TASK_REFERENCE = (
      'See more details about the task and its versions in '
      'https://github.com/rail-berkeley/d4rl/wiki/Tasks#antmaze'
  )

  # pylint: disable=protected-access
  # Step metadata fields.
  QPOS = dataset_builder._QPOS
  QVEL = dataset_builder._QVEL
  GOAL = dataset_builder._GOAL
  # pylint: enable=protected-access

  # pylint: disable=unexpected-keyword-arg,line-too-long
  BUILDER_CONFIGS = [
      # v0
      dataset_builder.BuilderConfig(
          name='umaze-v0',
          dataset_dir='ant_maze_new',
          env='mujoco',
          file_suffix=(
              'Ant_maze_u-maze_noisy_multistart_False_multigoal_False_sparse'
          ),
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='umaze-diverse-v0',
          dataset_dir='ant_maze_new',
          env='mujoco',
          file_suffix=(
              'Ant_maze_u-maze_noisy_multistart_True_multigoal_True_sparse'
          ),
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='medium-play-v0',
          dataset_dir='ant_maze_new',
          env='mujoco',
          file_suffix=(
              'Ant_maze_big-maze_noisy_multistart_True_multigoal_False_sparse'
          ),
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='medium-diverse-v0',
          dataset_dir='ant_maze_new',
          env='mujoco',
          file_suffix=(
              'Ant_maze_big-maze_noisy_multistart_True_multigoal_True_sparse'
          ),
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='large-diverse-v0',
          dataset_dir='ant_maze_new',
          env='mujoco',
          file_suffix='Ant_maze_hardest-maze_noisy_multistart_True_multigoal_True_sparse',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='large-play-v0',
          dataset_dir='ant_maze_new',
          env='mujoco',
          file_suffix='Ant_maze_hardest-maze_noisy_multistart_True_multigoal_False_sparse',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      # v2
      dataset_builder.BuilderConfig(
          name='umaze-v2',
          dataset_dir='ant_maze_v2',
          env='mujoco',
          file_suffix='Ant_maze_u-maze_noisy_multistart_False_multigoal_False_sparse_fixed',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='umaze-diverse-v2',
          dataset_dir='ant_maze_v2',
          env='mujoco',
          file_suffix='Ant_maze_u-maze_noisy_multistart_True_multigoal_True_sparse_fixed',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='medium-play-v2',
          dataset_dir='ant_maze_v2',
          env='mujoco',
          file_suffix='Ant_maze_big-maze_noisy_multistart_True_multigoal_False_sparse_fixed',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='medium-diverse-v2',
          dataset_dir='ant_maze_v2',
          env='mujoco',
          file_suffix='Ant_maze_big-maze_noisy_multistart_True_multigoal_True_sparse_fixed',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='large-diverse-v2',
          dataset_dir='ant_maze_v2',
          env='mujoco',
          file_suffix='Ant_maze_hardest-maze_noisy_multistart_True_multigoal_True_sparse_fixed',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
      dataset_builder.BuilderConfig(
          name='large-play-v2',
          dataset_dir='ant_maze_v2',
          env='mujoco',
          file_suffix='Ant_maze_hardest-maze_noisy_multistart_True_multigoal_False_sparse_fixed',
          step_metadata_keys=frozenset([QPOS, QVEL, GOAL]),
          description=_TASK_REFERENCE,
      ),
  ]
  # pylint: enable=unexpected-keyword-arg,line-too-long

  def __init__(self, **kwargs: Any):
    config = dataset_builder.DatasetConfig(
        name='antmaze',
        obs_len=29,
        action_len=8,
        qpos_len=15,
        qvel_len=14,
        goal_len=2,
    )
    super().__init__(ds_config=config, **kwargs)

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    ds_dir = self.builder_config.dataset_dir
    ds_name = self.builder_config.file_suffix + '.hdf5'
    path = dl_manager.download_and_extract(
        {
            'file_path': (
                'http://rail.eecs.berkeley.edu/datasets/offline_rl/'
                + ds_dir
                + '/'
                + ds_name
            )
        }
    )
    return {
        'train': self._generate_examples(path),
    }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""squad_question_generation dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.squad_question_generation import squad_question_generation


class SquadQuestionGenerationSplitDuTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for squad_question_generation (split_du) dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ["split_du"]
  DATASET_CLASS = squad_question_generation.SquadQuestionGeneration
  SPLITS = {
      "train": 3,  # Number of fake train example
      "validation": 1,  # Number of fake dev example
      "test": 1,  # Number of fake test example
  }
  DL_EXTRACT_RESULT = {
      "train": "split_du/train.json",
      "dev": "split_du/dev.json",
      "test": "split_du/test.json",
  }


class SquadQuestionGenerationSplitZhouTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for squad_question_generation (split_zhou) dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ["split_zhou"]
  DATASET_CLASS = squad_question_generation.SquadQuestionGeneration
  SPLITS = {
      "train": 3,  # Number of fake train example
      "validation": 1,  # Number of fake dev example
      "test": 1,  # Number of fake test example
  }
  DL_EXTRACT_RESULT = {
      "train": "split_zhou/train-v1.1.json",
      "dev": "split_zhou/dev-v1.1.json",
      "mapping": "split_zhou/",
      "redistribute": "split_zhou/",
  }


if __name__ == "__main__":
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gov_report dataset."""

import json

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.summarization.gov_report import gov_report

_REPORT = r"""
{
  "section_title":"",
  "paragraphs":[
  ],
  "subsections":[
    {
      "section_title":"Introduction",
      "paragraphs":[
        "p1",
        "p2"
      ],
      "subsections":[
        {
          "section_title":"Background",
          "paragraphs":[
            "p3"
          ],
          "subsections":[
          ]
        }
      ]
    },
    {
      "section_title":"Conclusion",
      "paragraphs":[
        "p6"
      ],
      "subsections":[
      ]
    }
  ]
}
"""


class GovReportTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gov_report dataset."""

  DATASET_CLASS = gov_report.GovReport
  SPLITS = {
      tfds.Split.TRAIN: 1,
      tfds.Split.VALIDATION: 1,
      tfds.Split.TEST: 1,
  }

  def test_flatten_structures_whitespace(self):
    self.assertEqual(
        gov_report._flatten_structure(json.loads(_REPORT), " ", 1, False),
        " Introduction p1 p2 Background p3 Conclusion p6",
    )

  def test_flatten_structures_html(self):
    self.assertEqual(
        gov_report._flatten_structure(json.loads(_REPORT), "\n", 1, True),
        "<h1></h1>\n<h2>Introduction</h2>\np1\np2\n<h3>Background</h3>\np3\n<h2>Conclusion</h2>\np6",
    )


if __name__ == "__main__":
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for file_utils."""

import os
import time
from unittest import mock

from absl.testing import flagsaver
from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import read_config


def test_default_data_dir():
  data_dir = file_utils.get_default_data_dir(given_data_dir=None)
  assert data_dir
  assert isinstance(data_dir, str)


def test_list_dataset_variants_with_configs(mock_fs: testing.MockFs):
  data_dir = epath.Path('/a')
  dataset_dir = data_dir / 'ds'
  configs_and_versions = {
      'x': ['1.0.0', '1.0.1'],
      'y': ['2.0.0'],
  }
  for config, versions in configs_and_versions.items():
    for version in versions:
      mock_fs.add_file(dataset_dir / config / version / 'dataset_info.json')
      mock_fs.add_file(dataset_dir / config / version / 'features.json')

  references = sorted(
      file_utils.list_dataset_variants(
          dataset_name='my_ds', dataset_dir=dataset_dir
      )
  )
  assert references == [
      naming.DatasetReference(
          dataset_name='my_ds', config='x', version='1.0.0', data_dir=data_dir
      ),
      naming.DatasetReference(
          dataset_name='my_ds', config='x', version='1.0.1', data_dir=data_dir
      ),
      naming.DatasetReference(
          dataset_name='my_ds', config='y', version='2.0.0', data_dir=data_dir
      ),
  ]


def test_list_dataset_variants_with_configs_no_versions(
    mock_fs: testing.MockFs,
):
  data_dir = epath.Path('/a')
  dataset_dir = data_dir / 'ds'
  configs_and_versions = {
      'x': ['1.0.0', '1.0.1'],
      'y': ['2.0.0'],
  }
  for config, versions in configs_and_versions.items():
    for version in versions:
      mock_fs.add_file(dataset_dir / config / version / 'dataset_info.json')
      mock_fs.add_file(dataset_dir / config / version / 'features.json')

  references = sorted(
      file_utils.list_dataset_variants(
          dataset_name='my_ds', dataset_dir=dataset_dir, include_versions=False
      )
  )
  assert references == [
      naming.DatasetReference(
          dataset_name='my_ds', config='x', data_dir=data_dir
      ),
      naming.DatasetReference(
          dataset_name='my_ds', config='y', data_dir=data_dir
      ),
  ]


def test_list_dataset_variants_without_configs(mock_fs: testing.MockFs):
  data_dir = epath.Path('/a')
  dataset_dir = data_dir / 'ds'
  # Version 1.0.0 doesn't have features.json, because it was generated with an
  # old version of TFDS.
  mock_fs.add_file(dataset_dir / '1.0.0' / 'dataset_info.json')
  mock_fs.add_file(dataset_dir / '1.0.1' / 'dataset_info.json')
  mock_fs.add_file(dataset_dir / '1.0.1' / 'features.json')

  # List dirs including datasets generated by old TFDS versions.
  references = sorted(
      file_utils.list_dataset_variants(
          dataset_name='my_ds',
          dataset_dir=dataset_dir,
          include_versions=True,
          include_old_tfds_version=True,
      )
  )
  assert references == [
      naming.DatasetReference(
          dataset_name='my_ds', version='1.0.0', data_dir=data_dir
      ),
      naming.DatasetReference(
          dataset_name='my_ds', version='1.0.1', data_dir=data_dir
      ),
  ]

  # List dirs excluding datasets generated by old TFDS versions.
  references = sorted(
      file_utils.list_dataset_variants(
          dataset_name='my_ds',
          dataset_dir=dataset_dir,
          include_versions=True,
          include_old_tfds_version=False,
      )
  )
  assert references == [
      naming.DatasetReference(
          dataset_name='my_ds', version='1.0.1', data_dir=data_dir
      )
  ]


def test_list_datasets_in_data_dir(mock_fs: testing.MockFs):
  data_dir = epath.Path('/a')
  mock_fs.add_file(data_dir / 'ds1/config1/1.0.0/dataset_info.json')
  mock_fs.add_file(data_dir / 'ds1/config1/1.0.0/features.json')
  mock_fs.add_file(data_dir / 'ds1/config1/2.0.0/dataset_info.json')
  mock_fs.add_file(data_dir / 'ds1/config1/2.0.0/features.json')
  mock_fs.add_file(data_dir / 'ds1/config2/1.0.0/dataset_info.json')
  mock_fs.add_file(data_dir / 'ds1/config2/1.0.0/features.json')
  mock_fs.add_file(data_dir / 'ds2/1.0.0/dataset_info.json')
  mock_fs.add_file(data_dir / 'ds2/1.0.0/features.json')

  # The following are problematic and should thus be ignored.
  mock_fs.add_file(
      os.path.join(data_dir, 'invalid-name/1.0.0/features.json'), content='x'
  )
  mock_fs.add_file(
      os.path.join(data_dir, 'invalid_version1/1.a.b/features.json'),
      content='x',
  )
  mock_fs.add_file(
      os.path.join(data_dir, 'invalid_version2/1.2.3.4/features.json'),
      content='x',
  )

  references = sorted(
      file_utils.list_datasets_in_data_dir(data_dir=epath.Path(data_dir))
  )
  data_dir = epath.Path('/a')
  assert references == [
      naming.DatasetReference(
          dataset_name='ds1',
          config='config1',
          version='1.0.0',
          data_dir=data_dir,
      ),
      naming.DatasetReference(
          dataset_name='ds1',
          config='config1',
          version='2.0.0',
          data_dir=data_dir,
      ),
      naming.DatasetReference(
          dataset_name='ds1',
          config='config2',
          version='1.0.0',
          data_dir=data_dir,
      ),
      naming.DatasetReference(
          dataset_name='ds2', version='1.0.0', data_dir=data_dir
      ),
  ]


def test_list_datasets_in_data_dir_with_namespace(mock_fs: testing.MockFs):
  namespace = 'ns'
  data_dir = epath.Path('/a')
  mock_fs.add_file(data_dir / 'ds1/config1/1.0.0/dataset_info.json')
  mock_fs.add_file(data_dir / 'ds1/config1/1.0.0/features.json')

  references = sorted(
      file_utils.list_datasets_in_data_dir(
          data_dir=epath.Path(data_dir),
          namespace=namespace,
          include_configs=True,
          include_versions=True,
      )
  )
  data_dir = epath.Path('/a')
  assert references == [
      naming.DatasetReference(
          dataset_name='ds1',
          namespace=namespace,
          config='config1',
          version='1.0.0',
          data_dir=data_dir,
      ),
  ]


@pytest.mark.parametrize(
    ['filename', 'result'],
    [
        ('abc', False),
        ('dataset_info.json', True),
        ('features.json', True),
        ('mnist-test.tfrecord-00000-of-00001', True),
        ('mnist-test.arrayrecord-00000-of-00001', True),
    ],
)
def test_looks_like_a_tfds_file(filename, result):
  assert file_utils._looks_like_a_tfds_file(filename) == result


@pytest.mark.parametrize(
    ['path', 'glob_result', 'expected'],
    [
        ('/a/*', ['/a/b', '/a/c'], ['/a/b', '/a/c']),
        ('/a/b', None, ['/a/b']),
        ('a/*', None, ['a/*']),
        ('/a/b@*', None, ['/a/b@*']),
    ],
)
def test_expand_glob(path, glob_result, expected):
  with mock.patch.object(epath, 'Path') as mock_epath:
    mock_epath.return_value.expanduser.return_value = path
    mock_epath.return_value.glob.return_value = glob_result
    actual = file_utils.expand_glob(path)
    if glob_result is not None:
      mock_epath.return_value.glob.assert_called_once_with(path[1:])
    else:
      mock_epath.return_value.glob.assert_not_called()
    actual = [os.fspath(p) for p in actual]
    assert actual == expected

if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""d4rl_adroit_relocate dataset."""
from typing import Any

from tensorflow_datasets.d4rl import dataset_builder
import tensorflow_datasets.public_api as tfds


class D4rlAdroitRelocate(dataset_builder.D4RLDatasetBuilder):
  """DatasetBuilder for d4rl_adroit_relocate dataset."""

  VERSION = tfds.core.Version('1.1.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.1.0': 'Added is_last.',
  }

  BUILDER_CONFIGS = dataset_builder.ADROIT_BUILDER_CONFIGS

  def __init__(self, **kwargs: Any):
    config = dataset_builder.DatasetConfig(
        name='relocate', obs_len=39, action_len=30, qpos_len=36, qvel_len=36
    )
    super().__init__(ds_config=config, **kwargs)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Xtreme dataset collection."""
import collections

from typing import Mapping

from tensorflow_datasets.core import dataset_collection_builder
from tensorflow_datasets.core import naming


class Xtreme(dataset_collection_builder.DatasetCollection):
  """Xtreme collection."""

  @property
  def info(self) -> dataset_collection_builder.DatasetCollectionInfo:
    return dataset_collection_builder.DatasetCollectionInfo.from_cls(
        dataset_collection_class=self.__class__,
        release_notes={
            "1.0.0": "Initial release",
        },
        homepage="https://sites.research.google/xtreme",
    )

  @property
  def datasets(
      self,
  ) -> Mapping[str, Mapping[str, naming.DatasetReference]]:
    return collections.OrderedDict(
        {
            "1.0.0": naming.references_for({
                "xnli": "xtreme_xnli:1.1.0",
                "pawsx": "xtreme_pawsx:1.0.0",
                "pos": "xtreme_pos:1.0.0",
                "ner": "wikiann:1.0.0",
                "xquad": "xquad:3.0.0",
                "mlqa": "mlqa:1.0.0",
                "tydiqa": "tydi_qa:3.0.0",
                "bucc": "bucc:1.0.0",
                "tatoeba": "tatoeba:1.0.0",
            })
        }
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""C4 dataset based on Common Crawl."""

import collections
import functools
import json
import os
from typing import Optional, Sequence
import uuid

from absl import logging
from etils import epath
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text import c4_utils

PageFeatures = c4_utils.PageFeatures

_DESCRIPTION = """\
A colossal, cleaned version of Common Crawl's web crawl corpus.

Based on Common Crawl dataset: https://commoncrawl.org

To generate this dataset, please follow
[the instructions from t5](https://github.com/google-research/text-to-text-transfer-transformer#c4).

Due to the overhead of cleaning the dataset, it is recommend you prepare it with
a distributed service like Cloud Dataflow. More info at
https://www.tensorflow.org/datasets/beam_datasets.
"""
_CITATION = """
@article{2019t5,
  author = {Colin Raffel and Noam Shazeer and Adam Roberts and Katherine Lee and Sharan Narang and Michael Matena and Yanqi Zhou and Wei Li and Peter J. Liu},
  title = {Exploring the Limits of Transfer Learning with a Unified Text-to-Text Transformer},
  journal = {arXiv e-prints},
  year = {2019},
  archivePrefix = {arXiv},
  eprint = {1910.10683},
}
"""
_VERSION = tfds.core.Version("3.1.0")

# TODO(adarob): Remove supported versions. Starting with 3.0.0, all generated
# datasets are automatically forward compatible. For example,
# tfds.load('c4:3.0.0') works even if the code is at 3.0.1.
_SUPPORTED_VERSIONS = [
    tfds.core.Version("2.3.1"),
    tfds.core.Version("2.3.0"),
    tfds.core.Version("2.2.1"),
    tfds.core.Version("2.2.0"),
]
RELEASE_NOTES = {
    "3.1.0": (
        "All: Select newest timestamp when deduping by URL (vs random); escape "
        "badword regexes."
        "Multilingual: Expand complete list of CC dumps (as of Aug 17, 2022); "
        "use soft badword filtering (99.9%); do not require spaces for "
        "Japanese when matching bad words (similar to Thai and Chinese); "
        "up minimum language detection threshold to 0.95 (from 0.7); allow "
        "badwords that filter >10% of documents for a given non-spaced "
        "language due to being common subwords."
    ),
    "3.0.1": "Remove mC4 languages with less than 10k pages.",
    "3.0.0": "Add multilingual version (mC4). Deterministic URL deduplication.",
    "2.3.1": "Hashing change.",
    "2.3.0": "Deduplicate lines within a page.",
    "2.2.1": "Update dataset_info.json",
}

_DOWNLOAD_HOST = "https://data.commoncrawl.org"
_WET_PATH_URL = (
    "https://data.commoncrawl.org/crawl-data/CC-MAIN-{cc_version}/wet.paths.gz"
)
_REALNEWS_DOMAINS_URL = "https://raw.githubusercontent.com/rowanz/grover/38f7184bd87237ae2d3bc330b99f1e2e246f6d51/realnews/domain_to_allowed_subdomains.json"
_OPENWEBTEXT_URLS_ZIP = "OpenWebText.zip"
_OPENWEBTEXT_URLS_URL = "https://mega.nz/#F!EZZD0YwJ!9_PlEQzdMVLaNdKv_ICNVQ"
_OPENWEBTEXT_URLS_FILE_PATTERN = "OpenWebText/Version 1/URLs/*.txt"
_EN_BADWORDS_URL = "https://raw.githubusercontent.com/LDNOOBW/List-of-Dirty-Naughty-Obscene-and-Otherwise-Bad-Words/25e679f03d96baa721cde20db9944649e8d0a844/en"
_BADWORDS_URL = "https://raw.githubusercontent.com/LDNOOBW/List-of-Dirty-Naughty-Obscene-and-Otherwise-Bad-Words/5faf2ba42d7b1c0977169ec3611df25a3c08eb13/{lang}"
_BADWORDS_LANGS = [
    "ar",
    "cs",
    "da",
    "de",
    "en",
    "eo",
    "es",
    "fa",
    "fi",
    "fil",
    "fr",
    "fr-CA-u-sd-caqc",
    "hi",
    "hu",
    "it",
    "ja",
    "kab",
    "ko",
    "nl",
    "no",
    "pl",
    "pt",
    "ru",
    "sv",
    "th",
    "tlh",
    "tr",
    "zh",
]
# Words that are allowed since they are common subwords in languages without
# spaces. These each filter >10% of documents of their language when disallowed.
_BADWORDS_ALLOWLIST = {"ja": {"sm", "グロ", "女の子"}, "zh": {"性"}}

DEFAULT_CC_VERSION = "2019-18"

OPENWEBTEXT_CC_VERSIONS = (  # August 2018 - July 2019
    "2019-18",  # Original default for single-crawl dataset (April 2019).
    "2019-30",
    "2019-26",
    "2019-22",
    "2019-13",
    "2019-09",
    "2019-04",
    "2018-51",
    "2018-47",
    "2018-43",
    "2018-39",
    "2018-34",
)

ALL_CC_VERSIONS = (  # as of August 1, 2022
    "2013-20",
    "2013-48",
    "2014-10",
    "2014-15",
    "2014-23",
    "2014-35",
    "2014-41",
    "2014-42",
    "2014-49",
    "2014-52",
    "2015-06",
    "2015-11",
    "2015-14",
    "2015-18",
    "2015-22",
    "2015-27",
    "2015-32",
    "2015-35",
    "2015-40",
    "2015-48",
    "2016-07",
    "2016-18",
    "2016-22",
    "2016-26",
    "2016-30",
    "2016-36",
    "2016-40",
    "2016-44",
    "2016-50",
    "2017-04",
    "2017-09",
    "2017-13",
    "2017-17",
    "2017-22",
    "2017-26",
    "2017-30",
    "2017-34",
    "2017-39",
    "2017-43",
    "2017-47",
    "2017-51",
    "2018-05",
    "2018-09",
    "2018-13",
    "2018-17",
    "2018-22",
    "2018-26",
    "2018-30",
    "2018-34",
    "2018-39",
    "2018-43",
    "2018-47",
    "2018-51",
    "2019-04",
    "2019-09",
    "2019-13",
    "2019-18",
    "2019-22",
    "2019-26",
    "2019-30",
    "2019-35",
    "2019-39",
    "2019-43",
    "2019-47",
    "2019-51",
    "2020-05",
    "2020-10",
    "2020-16",
    "2020-24",
    "2020-29",
    "2020-34",
    "2020-40",
    "2020-45",
    "2020-50",
    "2021-04",
    "2021-10",
    "2021-17",
    "2021-21",
    "2021-25",
    "2021-31",
    "2021-39",
    "2021-43",
    "2021-49",
    "2022-05",
    "2022-21",
    "2022-27",
)

_KNOWN_CORRUPT_WET_FILES = (  # as of August 12, 2022
    # files that raise UnicodeDecodeError
    "crawl-data/CC-MAIN-2017-13/segments/1490218203536.73/wet/CC-MAIN-20170322213003-00052-ip-10-233-31-227.ec2.internal.warc.wet.gz",
)

# Limited to languages in CLD3 that produce at least 10k pages when using the
# "multilingual" config below.
MC4_LANGUAGES = [
    "af",
    "am",
    "ar",
    "az",
    "be",
    "bg",
    "bg-Latn",
    "bn",
    "ca",
    "ceb",
    "co",
    "cs",
    "cy",
    "da",
    "de",
    "el",
    "el-Latn",
    "en",
    "eo",
    "es",
    "et",
    "eu",
    "fa",
    "fi",
    "fil",
    "fr",
    "fy",
    "ga",
    "gd",
    "gl",
    "gu",
    "ha",
    "haw",
    "hi",
    "hi-Latn",
    "hmn",
    "ht",
    "hu",
    "hy",
    "id",
    "ig",
    "is",
    "it",
    "iw",
    "ja",
    "ja-Latn",
    "jv",
    "ka",
    "kk",
    "km",
    "kn",
    "ko",
    "ku",
    "ky",
    "la",
    "lb",
    "lo",
    "lt",
    "lv",
    "mg",
    "mi",
    "mk",
    "ml",
    "mn",
    "mr",
    "ms",
    "mt",
    "my",
    "ne",
    "nl",
    "no",
    "ny",
    "pa",
    "pl",
    "ps",
    "pt",
    "ro",
    "ru",
    "ru-Latn",
    "sd",
    "si",
    "sk",
    "sl",
    "sm",
    "sn",
    "so",
    "sq",
    "sr",
    "st",
    "su",
    "sv",
    "sw",
    "ta",
    "te",
    "tg",
    "th",
    "tr",
    "uk",
    "ur",
    "uz",
    "vi",
    "xh",
    "yi",
    "yo",
    "zh",
    "zh-Latn",
    "zu",
]


class C4Config(tfds.core.BuilderConfig):
  """BuilderConfig for C4 dataset."""

  def __init__(
      self,
      name: str,
      languages: Sequence[str],
      cc_versions: Optional[Sequence[str]] = None,
      clean: bool = False,
      badwords_filter_fraction: float = 0.0,
      paragraph_filter: bool = False,
      dedupe: bool = True,
      realnewslike: bool = False,
      webtextlike: bool = False,
      **kwargs,
  ):
    """BuilderConfig for C4.

    Args:
      name: string, the name for the config.
      languages: list(string), the language code(s) to include.
      cc_versions: tuple(string), a collection of versions of Common Crawl to
        use as the raw source text. Set to None to use default.
      clean: bool, whether to heuristically filter out lines and pages
        considered low quality. Note: only expected to work reliably for English
        pages.
      badwords_filter_fraction: float, what fraction of pages to filter out that
        contain "bad words".
      paragraph_filter: bool, whether to filter out pages with too few or too
        short paragraphs.
      dedupe: bool, whether to deduplicate the dataset by paragraphs.
      realnewslike: bool, whether to limit to news domains as compiled by
        RealNews.
      webtextlike: bool, whether to limit to WebText-like URLs.
      **kwargs: keyword arguments forwarded to super.
    """
    super(C4Config, self).__init__(
        name=name,
        version=_VERSION,
        supported_versions=_SUPPORTED_VERSIONS,
        **kwargs,
    )

    if clean and tuple(languages) != ("en",):
      logging.warning(
          "C4 cleaning is only expected to work reliably for English pages."
      )

    if not 0.0 <= badwords_filter_fraction <= 1.0:
      raise ValueError(
          "`badwords_filter_fraction` must be between 0.0 and 1.0."
      )

    self.languages = languages
    self.cc_versions = cc_versions or (DEFAULT_CC_VERSION,)
    self.clean = clean
    self.badwords_filter_fraction = badwords_filter_fraction
    self.paragraph_filter = paragraph_filter
    self.dedupe = dedupe
    self.realnewslike = realnewslike
    self.webtextlike = webtextlike


class C4(tfds.core.BeamBasedBuilder):
  """C4 dataset based on Common Crawl."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = f"""
  You are using a C4 config that requires some files to be manually downloaded.
  For `c4/webtextlike`, download {_OPENWEBTEXT_URLS_ZIP} from
  {_OPENWEBTEXT_URLS_URL}.
  """

  BUILDER_CONFIGS = [
      C4Config(
          "en",
          languages=["en"],
          clean=True,
          dedupe=True,
          badwords_filter_fraction=1.0,
          description="English C4 dataset.",
      ),
      C4Config(
          "en.noclean",
          languages=["en"],
          clean=False,
          dedupe=False,
          badwords_filter_fraction=0.0,
          description=(
              "Disables all cleaning (deduplication, removal based on bad"
              " words, etc.)"
          ),
      ),
      C4Config(
          "realnewslike",
          languages=["en"],
          realnewslike=True,
          clean=True,
          dedupe=True,
          badwords_filter_fraction=1.0,
          description=(
              "Filters from the default config to only include content from the"
              " domains used in the 'RealNews' dataset (Zellers et al., 2019)."
          ),
      ),
      C4Config(
          "webtextlike",
          languages=["en"],
          cc_versions=OPENWEBTEXT_CC_VERSIONS,
          webtextlike=True,
          clean=True,
          dedupe=True,
          badwords_filter_fraction=1.0,
          description=(
              "Filters from the default config to only include content from the"
              " URLs in OpenWebText"
              " (https://github.com/jcpeterson/openwebtext)."
          ),
      ),
      C4Config(
          "multilingual",
          languages=MC4_LANGUAGES,
          cc_versions=ALL_CC_VERSIONS,
          clean=False,
          paragraph_filter=True,
          dedupe=True,
          badwords_filter_fraction=0.999,
          description=(
              "Multilingual C4 (mC4) has 101 languages and is generated from 86"
              " Common Crawl dumps."
          ),
      ),
  ]

  def _info(self):
    features = {
        "text": tfds.features.Text(),
        "url": tfds.features.Text(),
        "content-type": tfds.features.Text(),
        "content-length": tfds.features.Text(),
        "timestamp": tfds.features.Text(),
    }
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(features),
        citation=_CITATION,
        homepage="https://github.com/google-research/text-to-text-transfer-transformer#datasets",
    )

  def _split_generators(self, dl_manager, pipeline):
    # We will automatically download the first default CC version, but others
    # need to be manually downloaded.
    cc_versions = set(self.builder_config.cc_versions)
    files_to_download = {}
    files_to_download["wet_path_urls"] = [
        _WET_PATH_URL.format(cc_version=cc_version)
        for cc_version in cc_versions
    ]
    if self.builder_config.badwords_filter_fraction > 0.0:
      files_to_download["badwords"] = {
          lang: _BADWORDS_URL.format(lang=lang)
          for lang in _BADWORDS_LANGS
          if lang != "en"
      }
      # Use older "en" file for reproducibility of the original C4.
      files_to_download["badwords"]["en"] = _EN_BADWORDS_URL
    if self.builder_config.realnewslike:
      files_to_download["realnews_domains"] = _REALNEWS_DOMAINS_URL
    file_paths = dl_manager.download_and_extract(files_to_download)

    if self.builder_config.webtextlike:
      owt_path = os.path.join(dl_manager.manual_dir, _OPENWEBTEXT_URLS_ZIP)
      if not tf.io.gfile.exists(owt_path):
        raise AssertionError(
            "For the WebText-like config, you must manually download the "
            "following file from {0} and place it in {1}: {2}".format(
                _OPENWEBTEXT_URLS_URL,
                dl_manager.manual_dir,
                _OPENWEBTEXT_URLS_ZIP,
            )
        )
      file_paths["openwebtext_urls_zip"] = dl_manager.extract(owt_path)

    file_paths = tree_utils.map_structure(os.fspath, file_paths)

    pages_pcollection = self._get_pages_pcollection(
        pipeline, file_paths, dl_manager
    )

    def _filter(page, lang, predicate_fn):
      return page.language == lang and c4_utils.get_hashed_url_filter_fn(
          predicate_fn
      )(page)

    train_predicate_fn = lambda x: x % 1000 != 0  # 99.9%
    validation_predicate_fn = lambda x: x % 1000 == 0  # 00.1%

    if len(self.builder_config.languages) == 1:
      # Single-language version.
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TRAIN,
              gen_kwargs=dict(
                  split="train",
                  pages=pages_pcollection,
                  split_filter_fn=c4_utils.get_hashed_url_filter_fn(
                      predicate_fn=train_predicate_fn
                  ),
              ),
          ),
          tfds.core.SplitGenerator(
              name=tfds.Split.VALIDATION,
              gen_kwargs=dict(
                  split="validation",
                  pages=pages_pcollection,
                  split_filter_fn=c4_utils.get_hashed_url_filter_fn(
                      predicate_fn=validation_predicate_fn
                  ),
              ),
          ),
      ]

    splits = []
    for lang in self.builder_config.languages + [c4_utils.UNKNOWN_LANGUAGE]:
      splits.extend([
          tfds.core.SplitGenerator(
              name=lang,
              gen_kwargs=dict(
                  split=lang,
                  pages=pages_pcollection,
                  split_filter_fn=functools.partial(
                      _filter, lang=lang, predicate_fn=train_predicate_fn
                  ),
              ),
          ),
          tfds.core.SplitGenerator(
              name=f"{lang}-validation",
              gen_kwargs=dict(
                  split=f"{lang}-validation",
                  pages=pages_pcollection,
                  split_filter_fn=functools.partial(
                      _filter,
                      lang=lang,
                      predicate_fn=validation_predicate_fn,
                  ),
              ),
          ),
      ])
    return splits

  def _get_pages_pcollection(self, pipeline, file_paths, dl_manager):
    """Build PCollection of un-split page content."""
    beam = tfds.core.lazy_imports.apache_beam

    def download_wet_file(path, dl_dir):
      url = f"{_DOWNLOAD_HOST}/{path}"
      out_path = f"{dl_dir}/{path}"

      if tf.io.gfile.exists(out_path):
        c4_utils.get_counter_inc_fn("download_wet_url")("exists")
        return out_path

      tmp_dir = f"{out_path}.incomplete{uuid.uuid4().hex}"
      try:
        tf.io.gfile.makedirs(tmp_dir)
        downloader = tfds.download.download_manager.get_downloader()
        with downloader.tqdm():
          # TODO(slebedev): Investigate why pytype infers Promise[Future[...]].
          dl_path = downloader.download(url, tmp_dir).get().path  # type: ignore
        tf.io.gfile.rename(os.fspath(dl_path), out_path, overwrite=True)
      finally:
        if tf.io.gfile.exists(tmp_dir):
          tf.io.gfile.rmtree(tmp_dir)

        c4_utils.get_counter_inc_fn("download_wet_url")("downloaded")
      return out_path

    wet_file_paths = (
        pipeline
        | "create_wet_path_urls" >> beam.Create(file_paths["wet_path_urls"])
        | beam.io.ReadAllFromText(
            compression_type=beam.io.filesystem.CompressionTypes.UNCOMPRESSED
        )
        # Increase parallelism.
        | beam.Reshuffle()
        | "filter_corrupt_wet_files"
        >> beam.Filter(lambda p: p not in _KNOWN_CORRUPT_WET_FILES)
        | beam.Map(
            download_wet_file,
            dl_dir=os.path.join(dl_manager.download_dir, "c4_wet_files"),
        )
    )

    # Parse WET files and filter by length.
    # Output: [PageFeatures]
    pages = (
        wet_file_paths
        | beam.FlatMap(c4_utils.split_wet_file)
        | beam.Filter(c4_utils.is_valid_length)
    )

    # Optionally filter for RealNews domains.
    # Output: [PageFeatures]
    if self.builder_config.realnewslike:
      with tf.io.gfile.GFile(file_paths["realnews_domains"]) as f:
        realnews_domains = json.load(f)
      pages |= beam.Filter(c4_utils.is_realnews_domain, realnews_domains)

    # Normalize and deduplicate by URL.
    # Output: [PageFeatures]
    pages = (
        pages
        | "normalize_url" >> beam.Map(lambda p: (p.normalized_url, p))
        | "select_newest" >> beam.CombinePerKey(c4_utils.select_newest_page)
        | beam.Values()
    )

    # Optionally filter for WebText-like URLs.
    # Output: [PageFeatures]
    if self.builder_config.webtextlike:
      webtextlike_urls = (
          pipeline
          | "read_webtextlike_urls"
          >> beam.io.ReadFromText(
              os.path.join(
                  file_paths["openwebtext_urls_zip"],
                  _OPENWEBTEXT_URLS_FILE_PATTERN,
              )
          )
          | "add_dummy_page"
          >> beam.Map(lambda url: (c4_utils.normalize_url(url), PageFeatures()))
      )
      pages = (
          {
              "pages": pages | beam.Map(lambda p: (p.normalized_url, p)),
              "webtextlike_urls": webtextlike_urls,
          }
          | "group_webtextlike_urls" >> beam.CoGroupByKey()
          | beam.FlatMap(c4_utils.filter_by_webtextlike)
      )

    if self.builder_config.paragraph_filter:
      pages |= beam.Filter(c4_utils.paragraph_filter)

    if self.builder_config.clean:
      pages |= "clean_pages" >> beam.FlatMap(c4_utils.get_clean_page_fn())

    if self.builder_config.dedupe:
      pages = (
          # Also removes documents with too few sentences after deduplication.
          c4_utils.remove_duplicate_text(pages)  # pylint:disable=g-long-ternary
          if self.builder_config.clean
          else
          # If we are not cleaning, do not remove too-few-sentence documents.
          c4_utils.remove_duplicate_text(pages, min_num_sentences=0)
      )

    # Add detected language.
    if self.builder_config.languages == ["en"]:
      # Use langdetect for reproducibility of the original C4.
      pages |= beam.FlatMap(c4_utils.detect_english)
    else:
      pages = c4_utils.detect_languages(
          pages, valid_languages=self.builder_config.languages
      )

    if self.builder_config.badwords_filter_fraction > 0.0:
      # Create dictionary of badwords regex for each available language.
      badwords = collections.defaultdict(set)
      for lang, path in file_paths["badwords"].items():
        lang = lang.split("-")[0]  # remove suffix if present
        with epath.Path(path).open() as f:
          badwords[lang].update(l.strip() for l in f)

      for lang, allowlist in _BADWORDS_ALLOWLIST.items():
        badwords[lang] -= allowlist

      pages |= beam.Filter(
          c4_utils.get_badwords_filter_fn(
              badwords,
              filter_fraction=self.builder_config.badwords_filter_fraction,
          )
      )

    return pages

  def _build_pcollection(self, unused_pipeline, split, pages, split_filter_fn):
    beam = tfds.core.lazy_imports.apache_beam

    def _emit_examples(page: PageFeatures):
      c4_utils.get_counter_inc_fn(split)("examples")
      return page.url, {
          "url": page.url,
          "text": page.text,
          "content-type": page.content_type,
          "content-length": page.content_length,
          "timestamp": page.timestamp,
      }

    return pages | beam.Filter(split_filter_fn) | beam.Map(_emit_examples)
